"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Ontology Engineering Group
        http://www.oeg-upm.net/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2016 Ontology Engineering Group.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""
import hashlib
import logging
from abc import abstractmethod
from multiprocessing import Lock

from rdflib import URIRef

from agora.engine.fountain import onto as manager
from agora.engine.fountain.index import Index
from agora.engine.fountain.path import PathManager
from agora.engine.fountain.schema import Schema
from agora.engine.fountain.seed import SeedManager

__author__ = 'Fernando Serena'

log = logging.getLogger('agora.engine.fountain')


class FountainError(Exception):
    pass


class AbstractFountain(object):
    @abstractmethod
    def add_vocabulary(self, owl):
        # type: (str) -> iter
        raise NotImplementedError

    @abstractmethod
    def update_vocabulary(self, vid, owl):
        # type: (str, str) -> None
        raise NotImplementedError

    @abstractmethod
    def delete_vocabulary(self, vid):
        # type: (str) -> None
        raise NotImplementedError

    @abstractmethod
    def get_vocabulary(self, vid):
        # type: (str) -> str
        raise NotImplementedError

    @property
    @abstractmethod
    def vocabularies(self):
        raise NotImplementedError

    @property
    @abstractmethod
    def types(self):
        # type: () -> iter
        raise NotImplementedError

    @property
    @abstractmethod
    def properties(self):
        # type: () -> iter
        raise NotImplementedError

    @abstractmethod
    def get_type(self, type):
        raise NotImplementedError

    @abstractmethod
    def get_property(self, property):
        raise NotImplementedError

    @abstractmethod
    def get_paths(self, elm, force_seed=None):
        # type: (str, iter) -> (iter, iter)
        raise NotImplementedError

    @abstractmethod
    def add_seed(self, uri, type):
        # type: (str, str) -> str
        raise NotImplementedError

    @property
    @abstractmethod
    def prefixes(self):
        raise NotImplementedError

    @abstractmethod
    def add_prefixes(self, prefixes):
        raise NotImplementedError

    @property
    @abstractmethod
    def seeds(self):
        # type: () -> iter
        raise NotImplementedError

    @abstractmethod
    def get_seed(self, sid):
        # type: (str) -> dict
        raise NotImplementedError

    @abstractmethod
    def get_type_seeds(self, type):
        # type: (str) -> iter
        raise NotImplementedError

    @abstractmethod
    def delete_seed(self, sid):
        # type: (str) -> None
        raise NotImplementedError

    @abstractmethod
    def get_seed_type_digest(self, type):
        # type: (str) -> None
        raise NotImplementedError

    @abstractmethod
    def delete_type_seeds(self, type):
        # type: (str) -> None
        raise NotImplementedError

    @abstractmethod
    def connected(self, source, target):
        # type: (str, str) -> bool
        raise NotImplementedError


class Fountain(AbstractFountain):
    def __init__(self):
        self.__index = None
        self.__schema = None
        self.__pm = None
        self.__sm = None
        self.__lock = Lock()

    @property
    def index(self):
        # type: () -> Index
        return self.__index

    @index.setter
    def index(self, i):
        # type: (Index) -> None
        self.__index = i
        if self.__schema:
            self.__index.schema = self.__schema

    @property
    def seed_manager(self):
        # type: () -> SeedManager
        return self.__sm

    @seed_manager.setter
    def seed_manager(self, s):
        # type: (SeedManager) -> None
        self.__sm = s

    @property
    def schema(self):
        # type: () -> Schema
        return self.__schema

    @schema.setter
    def schema(self, s):
        # type: (Schema) -> None
        self.__schema = s
        if self.__index:
            self.__index.schema = self.__schema

    @property
    def path_manager(self):
        # type: () -> PathManager
        return self.__pm

    @path_manager.setter
    def path_manager(self, p):
        # type: (PathManager) -> None
        self.__pm = p

    def add_vocabulary(self, owl):
        # type: (str) -> iter
        with self.__lock:
            added_vocs_iter = manager.add_vocabulary(self.__schema, owl)
            if self.__schema.cache is not None:
                self.__schema.cache.stable = 0
            for vid in reversed(added_vocs_iter):
                self.__index.index_vocabulary(vid)
            self.__pm.calculate()
            self.__sm.validate()
            if self.__schema.cache is not None:
                self.__schema.cache.stable = 1
            return added_vocs_iter

    def update_vocabulary(self, vid, owl):
        # type: (str, str) -> None
        with self.__lock:
            manager.update_vocabulary(self.__schema, vid, owl)
            if self.__schema.cache is not None:
                self.__schema.cache.stable = 0
            for v in manager.get_vocabularies(self.__schema):
                self.__index.index_vocabulary(v)
            self.__pm.calculate()
            self.__sm.validate()
            if self.__schema.cache is not None:
                self.__schema.cache.stable = 1

    def delete_vocabulary(self, vid):
        # type: (str) -> None
        with self.__lock:
            manager.delete_vocabulary(self.__schema, vid)
            if self.__schema.cache is not None:
                self.__schema.cache.stable = 0
            self.__index.delete_vocabulary(vid)
            for v in manager.get_vocabularies(self.__schema):
                self.__index.index_vocabulary(v)
            self.__pm.calculate()
            self.__sm.validate()
            if self.__schema.cache is not None:
                self.__schema.cache.stable = 1

    def get_vocabulary(self, vid):
        # type: (str) -> str
        with self.__lock:
            return manager.get_vocabulary(self.__schema, vid)

    @property
    def vocabularies(self):
        return self.__schema.contexts

    @property
    def types(self):
        # type: () -> iter
        return self.__index.types

    @property
    def properties(self):
        # type: () -> iter
        return self.__index.properties

    def get_type(self, type):
        return self.__index.get_type(type)

    def get_property(self, property):
        return self.__index.get_property(property)

    def get_paths(self, elm, force_seed=None):
        # type: (str, iter) -> (iter, iter)
        return self.__pm.get_paths(elm, force_seed=force_seed)

    def connected(self, source, target):
        # type: (str, str) -> bool
        return self.__pm.are_connected(source, target)

    def add_seed(self, uri, type):
        # type: (str, str) -> str
        return self.__sm.add_seed(uri, type)

    @property
    def prefixes(self):
        return self.__index.schema.prefixes

    def add_prefixes(self, prefixes):
        current_ns = dict(self.__schema.graph.namespaces()).values()
        for prefix, ns in prefixes.items():
            if not (prefix.startswith('ns') and ns in current_ns):
                self.__schema.graph.namespace_manager.bind(prefix, URIRef(ns), replace=True, override=True)
        self.__schema.update_ns_dicts()
        if self.__schema.cache is not None:
            self.__schema.cache.stable = 0
        for vid in self.vocabularies:
            self.__index.index_vocabulary(vid)
        self.__pm.calculate()
        self.__sm.validate()
        if self.__schema.cache is not None:
            self.__schema.cache.stable = 1

    @property
    def seeds(self):
        # type: () -> iter
        return self.__sm.seeds

    def get_seed(self, sid):
        # type: (str) -> dict
        return self.__sm.get_seed(sid)

    def get_type_seeds(self, type):
        # type: (str) -> iter
        return self.__sm.get_type_seeds(type)

    def delete_seed(self, sid):
        self.__sm.delete_seed(sid)

    def delete_type_seeds(self, type):
        return self.__sm.delete_type_seeds(type)

    def get_seed_type_digest(self, type):
        m = hashlib.md5()
        for seed in sorted(self.__sm.get_type_seeds(type)):
            m.update(seed)
        return m.digest().encode('base64').strip()



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Ontology Engineering Group
        http://www.oeg-upm.net/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2016 Ontology Engineering Group.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""
import logging
from threading import Lock

from flask import request
from rdflib import URIRef
from shortuuid import uuid

from agora.engine.fountain import AbstractFountain
from agora.engine.fountain.onto import VocabularyNotFound, DuplicateVocabulary, VocabularyError
from agora.engine.fountain.seed import InvalidSeedError, DuplicateSeedError
from agora.server import Server, APIError, Conflict, TURTLE, NotFound, Client, HTML, tuples_force_seed, JSON, \
    dict_force_seed

__author__ = 'Fernando Serena'

log = logging.getLogger('agora.server.fountain')


def build(fountain, server=None, import_name=__name__):
    # type: (AbstractFountain, Server, str) -> Server

    if server is None:
        server = Server(import_name)

    @server.get('/seeds/id/<string:sid>')
    def get_seed(sid):
        try:
            return fountain.get_seed(sid)
        except InvalidSeedError, e:
            raise APIError(e.message)

    @server.delete('/seeds/<string:type>')
    def delete_type_seeds(type):
        try:
            fountain.delete_type_seeds(type)
        except TypeError as e:
            raise NotFound(e.message)

    @server.get('/prefixes')
    def prefixes():
        return fountain.prefixes

    @server.post('/prefixes')
    def add_prefixes(prefixes):
        fountain.add_prefixes(prefixes)

    @server.get('/types/<string:type>')
    def get_type(type):
        try:
            return fountain.get_type(type)
        except TypeError as e:
            raise NotFound(e.message)

    @server.get('/vocabs/<string:vid>', produce_types=(TURTLE, HTML))
    def get_vocabulary(vid):
        return fountain.get_vocabulary(vid)

    @server.get('/types')
    def types():
        return {'types': fountain.types}

    @server.delete('/seeds/id/<string:sid>')
    def delete_seed(sid):
        try:
            fountain.delete_seed(sid)
        except InvalidSeedError, e:
            raise NotFound(e.message)

    @server.post('/seeds')
    def add_seed(seed_dict):
        try:
            sid = fountain.add_seed(seed_dict['uri'], seed_dict['type'])
            return server.url_for('get_seed', sid=sid)
        except (TypeError, ValueError) as e:
            raise APIError(e.message)
        except DuplicateSeedError as e:
            raise Conflict(e.message)

    @server.get('/paths/<string:elm>')
    def get_paths(elm):
        force_seed = request.args.getlist('force_seed', None)
        force_seed = None if not force_seed else map(lambda x: ('http://{}'.format(uuid()), x), force_seed)
        try:
            return fountain.get_paths(elm, force_seed=force_seed)
        except TypeError, e:
            raise APIError(e.message)

    @server.post('/paths', produce_types=(JSON,), consume_types=(JSON,))
    def get_paths_force_seeds(req_json):
        force_seed = list(tuples_force_seed(req_json['force_seed']))
        try:
            return fountain.get_paths(req_json['elm'], force_seed=force_seed)
        except TypeError, e:
            raise APIError(e.message)

    @server.get('/seeds/<string:type>')
    def get_type_seeds(type):
        try:
            return fountain.get_type_seeds(type)
        except TypeError as e:
            raise NotFound(e.message)

    @server.get('/seeds/<string:type>/digest')
    def get_seed_type_digest(type):
        try:
            return {'digest': fountain.get_seed_type_digest(type)}
        except TypeError as e:
            raise NotFound(e.message)

    @server.delete('/vocabs/<string:vid>')
    def delete_vocabulary(vid):
        fountain.delete_vocabulary(vid)

    @server.get('/vocabs')
    def vocabularies():
        return fountain.vocabularies

    @server.get('/properties/<string:property>')
    def get_property(property):
        try:
            return fountain.get_property(property)
        except TypeError as e:
            raise NotFound(e.message)

    @server.post('/vocabs', consume_types=('text/turtle',))
    def add_vocabulary(owl):
        try:
            return fountain.add_vocabulary(owl)
        except VocabularyNotFound, e:
            raise APIError('Ontology URI not found: {}'.format(e.message))
        except DuplicateVocabulary, e:
            raise Conflict(e.message)
        except VocabularyError, e:
            raise APIError(e.message)

    @server.get('/seeds')
    def seeds():
        return fountain.seeds

    @server.get('/properties')
    def properties():
        return {'properties': fountain.properties}

    @server.put('/vocabs/<string:vid>', consume_types=('text/turtle',))
    def update_vocabulary(owl, vid):
        try:
            fountain.update_vocabulary(vid, owl)
        except VocabularyNotFound, e:
            raise APIError('Ontology URI not found: {}'.format(e.message))
        except VocabularyError, e:
            raise APIError(e.message)

    return server


class FountainClient(Client, AbstractFountain):
    def __init__(self, host='localhost', port=5000):
        super(FountainClient, self).__init__(host, port)
        self.__types = {}
        self.__properties = {}
        self.__prefixes = {}
        self.__lock = Lock()

    @property
    def properties(self):
        response = self._get_request('properties')
        return response.get('properties')

    def get_type_seeds(self, type):
        response = self._get_request('seeds/{}'.format(type))
        return response.get('seeds')

    def get_seed(self, sid):
        response = self._get_request('seeds/id/{}'.format(sid))
        return response

    @property
    def prefixes(self):
        prefixes_raw = self._get_request('prefixes')
        return {prefix: URIRef(prefixes_raw[prefix]) for prefix in prefixes_raw}

    def add_prefixes(self, prefixes):
        self._post_request('/prefixes', data=prefixes, content_type='application/json')

    def update_vocabulary(self, vid, owl):
        response = self._put_request('vocabs/{}'.format(vid), owl)
        return response

    def get_paths(self, elm, force_seed=None):
        if force_seed:
            url = 'paths'
            type_seed_dict = dict_force_seed(force_seed)
            req_json = {'elm': str(elm), 'force_seed': type_seed_dict}
            response = self._post_request(url, req_json, content_type='application/json', accept='application/json')
        else:
            url = 'paths/%s' % elm
            response = self._get_request(url, accept='application/json')

        return response

    @property
    def seeds(self):
        response = self._get_request('seeds')
        return response

    def delete_type_seeds(self, type):
        response = self._delete_request('seeds/{}'.format(type))
        return response

    def get_property(self, property):
        try:
            return self._get_request('properties/{}'.format(property))
        except IOError as e:
            raise TypeError(e.message['text'])

    def get_vocabulary(self, vid):
        return self._get_request('vocabs/{}'.format(vid), accept='text/turtle')

    def delete_vocabulary(self, vid):
        return self._delete_request('vocabs/{}'.format(vid))

    def get_type(self, type):
        try:
            return self._get_request('types/{}'.format(type))
        except IOError as e:
            raise TypeError(e.message['text'])

    @property
    def types(self):
        response = self._get_request('types')
        return response.get('types')

    @property
    def vocabularies(self):
        response = self._get_request('vocabs')
        return response

    def add_seed(self, uri, type):
        response = self._post_request('seeds',
                                      {'uri': uri,
                                       'type': type},
                                      content_type='application/json')
        return response

    def connected(self, source, target):
        url = 'paths/{}'.format(target)
        url += '?force_seed={}'.format(source)
        response = self._get_request(url)
        paths = response.get('paths', [])
        if len(paths) == 1:
            path = paths.pop()
            connected = path['cycles'] or path['steps']
        else:
            connected = len(paths) > 1
        return bool(connected)

    def delete_seed(self, sid):
        raise NotImplementedError

    def get_seed_type_digest(self, type):
        response = self._get_request('seeds/{}/digest'.format(type))
        return response

    def add_vocabulary(self, owl):
        try:
            response = self._post_request('vocabs', owl)
            return response
        except IOError as e:
            if e.message['code'] == 409:
                raise DuplicateVocabulary(e.message['text'])
            raise e


def client(host='localhost', port=5000):
    # type: (str, int) -> FountainClient
    return FountainClient(host, port)



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Ontology Engineering Group
        http://www.oeg-upm.net/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2016 Ontology Engineering Group.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""
import json
import logging
import os
import urlparse
from functools import wraps

import requests
from flask import Flask, jsonify, request, url_for, make_response
from flask import Response
from flask import stream_with_context
from flask_negotiate import consumes, produces
from rdflib import URIRef

__author__ = 'Fernando Serena'

JSON = 'application/json'
TURTLE = 'text/turtle'
HTML = 'text/html'

log = logging.getLogger('agora.server')


def tuples_force_seed(request_json):
    for type, seeds in request_json.items():
        for seed in seeds:
            yield (URIRef(seed), type)


def dict_force_seed(force_seed):
    fs_dict = {}
    for seed, type in force_seed:
        if type not in fs_dict:
            fs_dict[type] = []
        fs_dict[type].append(seed)
    return fs_dict


class APIError(Exception):
    status_code = 400

    def __init__(self, message, status_code=None, payload=None):
        Exception.__init__(self)
        self.message = str(message)
        if status_code is not None:
            self.status_code = status_code
        self.payload = payload

    def to_dict(self):
        rv = dict(self.payload or ())
        rv['message'] = self.message
        return rv


class NotFound(APIError):
    def __init__(self, message, payload=None):
        super(NotFound, self).__init__(message, 404, payload)


class Conflict(APIError):
    def __init__(self, message, payload=None):
        super(Conflict, self).__init__(message, 409, payload)


def handle_invalid_usage(error):
    response = jsonify(error.to_dict())
    response.status_code = error.status_code
    return response


class Server(Flask):
    def __init__(self, import_name, **kwargs):
        base = os.path.dirname(os.path.realpath(__file__))
        t_folder = base + '/templates'
        s_folder = base + '/static'
        super(Server, self).__init__(import_name, template_folder=t_folder, static_folder=s_folder)
        self.errorhandler(APIError)(handle_invalid_usage)

    def response(self, rv, content_type=JSON, status=200):
        if content_type == JSON:
            response = jsonify(rv)

        else:
            response = self.make_response(rv)
            response.headers['Content-Type'] = content_type
        response.status_code = status
        return response

    @staticmethod
    def url_for(func_name, **kwargs):
        return url_for(func_name, _external=True, **kwargs)

    def make_response(self, rv):
        return super(Server, self).make_response(rv)

    def produce(self, result, produce_types=()):
        if result is None:
            result = ''

        headers = None
        if type(result) == tuple:
            result, headers = result

        if hasattr(result, 'next'):
            response = Response(stream_with_context(result), mimetype=produce_types[0])
        elif JSON in produce_types:
            response = jsonify(result)
        else:
            response = self.make_response(str(result))
            response.headers['Content-Type'] = produce_types[0] if produce_types else 'text/plain'

        if headers:
            response.headers.extend(headers.items())

        return response

    @property
    def request_args(self):
        return request.args

    def get(self, rule, produce_types=(JSON, HTML,)):
        # type: (str, iter) -> callable
        def decorator(f):
            @wraps(f)
            def wrap(*args, **kwargs):
                response = self.produce(f(*args, **kwargs), produce_types)
                return response

            return self.route(rule, methods=['GET'])(produces(*produce_types)(wrap))

        return decorator

    def post(self, rule, consume_types=(JSON,), produce_types=()):
        # type: (str, iter, iter) -> callable
        def decorator(f):
            @wraps(f)
            def wrap(*args, **kwargs):
                data = request.data
                if request.content_type == JSON:
                    data = request.json
                # Request data is passed as first argument
                result = f(data, *args, **kwargs)
                response = self.produce(result, produce_types)
                if not produce_types:
                    response.headers['Location'] = result
                    response.status_code = 201
                else:
                    response.status_code = 200
                return response

            content_f = consumes(*consume_types)(wrap)
            if produce_types:
                content_f = produces(*produce_types)(content_f)
            return self.route(rule, methods=['POST'])(content_f)

        return decorator

    def put(self, rule, consume_types=(JSON,), produce_types=()):
        # type: (str, iter, iter) -> callable
        def decorator(f):
            @wraps(f)
            def wrap(*args, **kwargs):
                data = request.data
                if request.content_type == JSON:
                    data = request.json
                # Request data is passed as first argument
                result = f(data, *args, **kwargs)
                response = self.produce(result, produce_types)
                return response

            content_f = consumes(*consume_types)(wrap)
            if produce_types:
                content_f = produces(*produce_types)(content_f)
            return self.route(rule, methods=['PUT'])(content_f)

        return decorator

    def delete(self, rule, produce_types=()):
        # type: (str, iter) -> callable
        def decorator(f):
            @wraps(f)
            def wrap(*args, **kwargs):
                f(*args, **kwargs)
                response = make_response()
                response.status_code = 200
                return response

            return self.route(rule, methods=['DELETE'])(wrap)  # (produces(*produce_types)(wrap))

        return decorator


class Client(object):
    def __init__(self, host='localhost', port=9002):
        self.host = 'http://{}:{}'.format(host, port)

    def _get_request(self, path, accept='application/json'):
        try:
            stream = accept == 'application/agora-quad'
            response = requests.get(urlparse.urljoin(self.host, path, allow_fragments=True).replace('#', '%23'),
                                    headers={'Accept': accept},
                                    stream=stream)
            if response.status_code != 200:
                try:
                    message = response.json()['message']
                except (ValueError, KeyError):
                    message = response.content
                raise IOError({'code': response.status_code, 'text': message})
            if stream:
                return response.iter_lines()
            if accept == 'application/json':
                return response.json()
            return response.content
        except requests.ConnectionError:
            message = 'Remote service is not available'
            raise EnvironmentError(message)

    @staticmethod
    def __process_request_data(data, content_type):
        if content_type == JSON:
            return json.dumps(data)
        return str(data)

    def _post_request(self, path, data, content_type='text/turtle', accept='application/json'):
        response = requests.post(urlparse.urljoin(self.host, path),
                                 data=self.__process_request_data(data, content_type),
                                 headers={'Content-Type': content_type, 'Accept': accept})
        if response.status_code == 200:
            try:
                if accept == 'application/json':
                    return response.json()
            except ValueError:
                pass
            return response.content
        elif response.status_code == 201:
            return response.headers['Location']
        else:
            try:
                message = response.json()['message']
            except (ValueError, KeyError):
                message = response.content
            # log.warning(message)
            raise IOError({'code': response.status_code, 'text': message})

    def _put_request(self, path, data, content_type='text/turtle', accept='application/json'):
        response = requests.put(urlparse.urljoin(self.host, path),
                                data=self.__process_request_data(data, content_type),
                                headers={'Content-Type': content_type, 'Accept': accept})
        if response.status_code == 200:
            try:
                if accept == 'application/json':
                    return response.json()
            except ValueError:
                pass
            return response.content
        elif response.status_code == 201:
            return response.headers['Location']
        else:
            try:
                message = response.json()['message']
            except (ValueError, KeyError):
                message = response.content
            # log.warning(message)
            raise IOError({'code': response.status_code, 'text': message})

    def _delete_request(self, path):
        response = requests.delete(urlparse.urljoin(self.host, path))
        if response.status_code != 200:
            log.warning(response.content)
            raise IOError({'code': response.status_code, 'text': response.content})



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Ontology Engineering Group
        http://www.oeg-upm.net/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2016 Ontology Engineering Group.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""
import logging
import shutil
from threading import RLock

from rdflib import ConjunctiveGraph

from agora.engine.utils import prepare_store_path
from agora.engine.utils.cache import ContextGraph

__author__ = 'Fernando Serena'

log = logging.getLogger('agora.engine.utils.cache')


def get_cached_triple_store(cache, persist_mode=False, base='store', path='', **kwargs):
    if persist_mode:
        full_path = prepare_store_path(base, path)
        graph = ContextGraph(cache, 'Sleepycat')
        try:
            graph.open(full_path, create=True)
        except Exception as e:
            rmtree(full_path)
            raise EnvironmentError("Graph store is corrupted")

    else:
        graph = ContextGraph(cache)

    graph.store.graph_aware = False
    return graph


_stores = {}
_st_lock = RLock()


def get_triple_store(persist_mode=False, base='store', path='', **kwargs):
    if persist_mode:
        full_path = prepare_store_path(base, path)
        graph = ConjunctiveGraph('Sleepycat', identifier=path)
        try:
            graph.open(full_path, create=True)
        except Exception as e:
            rmtree(full_path)
            raise EnvironmentError(e.message)
    else:
        graph = ConjunctiveGraph()
        graph.store.graph_aware = False

    return graph


def rmtree(path):
    try:
        shutil.rmtree(path)
    except OSError:
        pass



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Ontology Engineering Group
        http://www.oeg-upm.net/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2017 Ontology Engineering Group.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""
from Queue import Queue, Empty
from threading import Thread

import networkx as nx
from rdflib import Variable
from rdflib.plugins.sparql.sparql import AlreadyBound
from rdflib.plugins.sparql.sparql import QueryContext

from agora.collector.execution import StopException
from agora.engine.plan import AGP

__author__ = 'Fernando Serena'


class Context(object):
    def __init__(self, **kwargs):
        self.map = kwargs

    def __setitem__(self, key, value):
        if isinstance(key, Variable):
            self.map[key] = value

    def __getitem__(self, item):
        return self.map.get(item, None)

    def __eq__(self, other):
        if isinstance(other, Context):
            return self.map.items() == other.map.items()
        return False

    def __contains__(self, item):
        return item in self.map

    def __repr__(self):
        s = ''
        for k in sorted(self.map.keys()):
            if isinstance(k, Variable):
                s += u'{}={} '.format(k.toPython(), self.map[k].toPython())
        return s

    @property
    def variables(self):
        return set([k for k in self.map.keys() if isinstance(k, Variable)])

    def issubset(self, other):
        for k in self.map.keys():
            if k not in other.map or self.map[k] != other.map[k]:
                return False

        return True


class ContextCollection(set):
    def __init__(self):
        super(ContextCollection, self).__init__()

    def __contains__(self, other):
        # type: (Context) -> bool
        for c in self:
            if c.map == other.map:
                return True
        return False


def __base_generator(agp, ctx, fragment):
    # type: (AGP, QueryContext, iter) -> iter
    tp_single_vars = [(tp, tp.s if isinstance(tp.s, Variable) else tp.o) for tp in agp if
                      isinstance(tp.s, Variable) != isinstance(tp.o, Variable)]

    wire = agp.wire
    ignore_tps = [str(tp) for (tp, v) in tp_single_vars if len(list(wire.neighbors(v))) > 1]
    for tp, ss, _, so in fragment:
        if ctx.stop is not None:
            if ctx.stop.isSet():
                break

        if str(tp) in ignore_tps:
            continue
        kwargs = {}
        if isinstance(tp.o, Variable):
            kwargs[tp.o] = so
        if isinstance(tp.s, Variable):
            kwargs[tp.s] = ss

        if kwargs:
            yield Context(**kwargs), tp


def __exploit(c1, c2):
    # type: (Context, Context) -> Context
    new_dict = {v: c1[v] for v in c1.map}
    for v in set.difference(set(c2.map.keys()), set(c1.map.keys())):
        new_dict[v] = c2[v]

    c = Context(**new_dict)
    return c


def __joins(c, x, v_paths):
    # type: (Context, Context) -> bool
    intersection = c.variables.intersection(x.variables)
    for v in intersection:
        if c[v] != x[v]:
            return False

    return bool(intersection)


def common_descendants(graph, x, c, base):
    if base:
        return False
    try:
        for dx in nx.descendants(graph, c):
            for dc in nx.descendants(graph, x):
                if dx.map == dc.map:
                    return True
    except Exception as e:
        raise e
    return False


def filter_successor(graph, v_paths, c, x, base):
    return c.map != x.map and __joins(c, x, v_paths) and not common_descendants(graph, c, x, base)


def union(x, y):
    r = ContextCollection()
    for c in x:
        r.add(c)
    for c in y:
        r.add(c)
    return r


def __eval_delta(c, graph, v_paths, roots, variables, base=False):
    # type: (Context, nx.DiGraph) -> iter

    solutions = ContextCollection()

    root_candidates = reduce(lambda x, y: union(x, set(graph.successors(y))), roots, set())
    root_candidates = filter(lambda x: set(x.variables).symmetric_difference(c.variables), root_candidates)
    for root in root_candidates:
        if filter_successor(graph, v_paths, c, root, base):
            inter = __exploit(root, c)

            if inter not in graph:
                graph.add_node(inter)
                graph.add_edge(root, inter)
                graph.add_edge(c, inter)
            else:
                continue

            if len(graph.nodes()) > 5000:
                break

            # print 'DELTA ({}) {}'.format(len(inter.variables), inter)
            if len(inter.variables) == len(variables):
                solutions.add(inter)
            else:
                pred = filter(lambda x: graph.out_degree(x) > 1, [root, c])
                for s in __eval_delta(inter, graph, v_paths, pred, variables):
                    solutions.add(s)

    return solutions


def __query_context(ctx, c):
    # type: (QueryContext, Context) -> QueryContext
    q = ctx.push()
    for k, v in c.map.items():
        try:
            q[k] = v
        except AlreadyBound:
            pass
    return q


def __generate(data):
    queue, agp, ctx, generator = data['queue'], data['agp'], data['context'], data['gen']
    try:
        for c, tp in __base_generator(agp, ctx, generator):
            queue.put((c, tp))
            if ctx.stop is not None:
                if ctx.stop.isSet():
                    break
    except StopException:
        if ctx.stop is not None:
            ctx.stop.set()
    finally:
        data['collecting'] = False


def incremental_eval_bgp(ctx, bgp):
    # type: (QueryContext, iter) -> iter

    fragment_generator = ctx.graph.gen(bgp, filters=ctx.filters, stop_event=ctx.stop, follow_cycles=ctx.follow_cycles)
    queue = Queue()
    if fragment_generator is not None:
        dgraph = nx.DiGraph()
        agp = ctx.graph.build_agp(bgp)

        variables = set([v for v in agp.wire.nodes() if isinstance(v, Variable)])
        dgraph.add_nodes_from(variables)

        wire = agp.wire
        roots = agp.roots
        non_roots = [node for node in wire.nodes() if isinstance(node, Variable) and node not in roots]
        v_paths = {}
        complex_agp = False

        for root in roots:
            for node in non_roots:
                if node not in v_paths:
                    v_paths[node] = set()
                    try:
                        node_paths = list(nx.all_simple_paths(wire, root, node))
                        v_paths[node] = node_paths
                        if len(node_paths) > 1:
                            complex_agp = True
                    except nx.NetworkXNoPath:
                        pass

        gen_data = {
            'queue': queue,
            'agp': agp,
            'context': ctx,
            'gen': fragment_generator,
            'collecting': True,
        }

        gen_thread = Thread(target=__generate, args=(gen_data,))
        gen_thread.start()

        try:
            if complex_agp:
                raise Exception

            while gen_data['collecting'] or not queue.empty():
                if ctx.stop is not None:
                    if ctx.stop.isSet():
                        raise StopIteration()
                try:
                    c, tp = queue.get(timeout=0.01)
                    if len(dgraph.nodes()) > 5000:
                        break
                    [dgraph.add_edge(v, c) for v in c.variables]

                    # print 'BASE ({}) {}'.format(len(c.variables), c)
                    if len(c.variables) == len(variables):
                        yield __query_context(ctx, c).solution()
                    else:
                        if isinstance(tp.o, Variable) and isinstance(tp.s, Variable):
                            for solution in __eval_delta(c, dgraph, v_paths, c.variables, variables, base=True):
                                yield __query_context(ctx, solution).solution()
                except Empty:
                    pass
        except KeyboardInterrupt as e:
            if ctx.stop is not None:
                ctx.stop.set()
            raise e
        finally:
            gen_thread.join()



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Ontology Engineering Group
        http://www.oeg-upm.net/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2016 Ontology Engineering Group.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""

import copy
import logging
import traceback
from abc import abstractmethod

import networkx as nx
from rdflib import RDF, URIRef, Variable, BNode, Graph

from agora.engine.fountain import AbstractFountain
from agora.engine.plan.agp import AGP, TP
from agora.engine.plan.graph import graph_plan, __extend_uri
from agora.engine.utils import Wrapper

__author__ = 'Fernando Serena'

log = logging.getLogger('agora.engine.plan')


def _extend_with_super(fountain, s):
    extension = set()
    for t in s:
        extension.add(t)
        extension.update(set(fountain.get_type(t)['super']))
    return extension


def find_root_types(fountain, root_tps, graph, extend=True):
    root_types = {}
    roots_dict = {}
    for (s, p, o) in root_tps:
        tp_types = set()
        if p == RDF.type:
            if not isinstance(o, URIRef):
                tp_types.update(set(fountain.types))
            else:
                t = str(graph.qname(o))
                tp_types.add(t)
                tp_types.update(set(fountain.get_type(t)['sub']))
        else:
            tp_types.update(set(fountain.get_property(str(graph.qname(p)))['domain']))

        tp_types = set(filter(
            lambda t: not set.intersection(set(fountain.get_type(t)['super']), tp_types) or fountain.get_type(t)[
                'spec_refs'], tp_types))
        roots_dict[(s, p, o)] = tp_types

    for (s, p, o) in roots_dict:
        root_join = [(sr, pr, orr) for (sr, pr, orr) in roots_dict if sr == s and (sr, pr, orr) != (s, p, o)]
        base_types = roots_dict[(s, p, o)]
        if extend:
            base_types = _extend_with_super(fountain, base_types)
        types = [base_types]
        for (sr, pr, orr) in root_join:
            join_types = roots_dict[(sr, pr, orr)]
            if extend:
                join_types = _extend_with_super(fountain, join_types)
            else:
                for jt in join_types.copy():
                    join_types.update(set.intersection(set(fountain.get_type(jt)['sub']), base_types))
            types.append(join_types)
        types = set.intersection(*types)
        if types:
            root_types[(s, p, o)] = types

    return root_types


def _validate_agp_context(fountain, c):
    s_join_tested = []
    for s, pr, o in c.triples((None, None, None)):
        if pr == RDF.type:
            continue
        if (s, pr, o) not in s_join_tested:
            s_join = [(x, pj, y) for (x, pj, y) in c.triples((s, None, None)) if pj != pr]
            root_types = find_root_types(fountain, s_join + [(s, pr, o)], c)
            s_join_tested.append((s, pr, o))
            if not root_types:
                raise TypeError('bad subject join')
        so_join = list(c.triples((o, None, None)))
        pr_dict = fountain.get_property(c.qname(pr))
        for (so, pro, oo) in so_join:
            if pro == RDF.type:
                if isinstance(so, URIRef):
                    pro_domain = {c.qname(oo)}
                else:
                    pro_domain = set(fountain.types)
            else:
                pro_domain = set(fountain.get_property(c.qname(pro))['domain'])
            if not set.intersection(pro_domain, set(pr_dict['range'])):
                raise TypeError('bad subject-object join')
        o_join = [(x, pj, y) for (x, pj, y) in c.triples((None, None, o)) if pj != pr]
        for (so, pro, oo) in o_join:
            if pro == RDF.type:
                pro_range = set(fountain.types)
            else:
                pro_range = set(fountain.get_property(c.qname(pro))['range'])
            if not set.intersection(pro_range, set(pr_dict['range'])):
                raise TypeError('bad object-object join')


def _chunks(l, n):
    for i in xrange(0, len(l), n):
        yield l[i:i + n]


def _build_root_paths(fountain, path, rt, root_path_id):
    def _continue(root, steps=None, index=0, last_pr=None, last_ty=None, last_pr_dict=None):
        if steps is None:
            steps = []

        if index >= len(path):
            yield steps
        else:
            index_pr = path[index]
            if index == 0:
                step = {'type': root, 'property': index_pr, 'root': root_path_id}
                steps.append(step)
                for p in _continue(root, steps, index=1, last_pr=index_pr, last_ty=root):
                    yield p
            else:
                if last_pr_dict is None:
                    last_pr_dict = fountain.get_property(last_pr)

                if index_pr != 'rdf:type':
                    pr_dict = fountain.get_property(index_pr)
                    last_pr_constraints = last_pr_dict['constraints']
                    expected = set.intersection(set(pr_dict['domain']), set(last_pr_dict['range']))
                    expected = filter(lambda x: not set.intersection(set(fountain.get_type(x)['super']), expected),
                                      expected)
                else:
                    expected = set(last_pr_dict['range'])  # no more iterations from here!
                    last_pr_constraints = []
                for et in expected:
                    if last_ty in last_pr_constraints:
                        if et not in last_pr_constraints[last_ty]:
                            continue
                    step = {'property': index_pr, 'type': et, 'root': root_path_id}
                    et_steps = steps[:]
                    et_steps.append(step)
                    for p in _continue(root, et_steps, index=index + 1, last_pr=index_pr, last_ty=et):
                        yield p

    for p in _continue(rt):
        yield p


def _tp_to_context(tp):
    def node(elm):
        if isinstance(elm, Variable):
            return BNode(elm.toPython())
        else:
            return elm

    return node(tp.s), node(tp.p), node(tp.o)


def _subject_transform(x):
    if isinstance(x, BNode):
        return Variable(x)
    return x


def _get_tp_paths(fountain, agp, force_seed=None):
    tp_paths = {}
    tp_hints = {}
    tp_cycles = {}
    type_dicts = {}

    wire = agp.wire
    roots = agp.roots
    graph = agp.graph

    context_force_seeds = {}

    try:
        map(lambda c: _validate_agp_context(fountain, c), graph.contexts())
    except TypeError:
        return tp_paths, tp_hints, tp_cycles

    agp_paths = {}
    bgp_paths = {}
    seed_paths = {}
    str_roots = map(lambda x: str(x), roots)
    for c in graph.contexts():
        root_tps = filter(lambda (s, pr, o): str(s).replace('?', '') in str_roots, c.triples((None, None, None)))

        root_types = find_root_types(fountain, root_tps, c, extend=True)
        root_types = {_subject_transform(tp[0]): types for tp, types in root_types.items()}
        spec_root_types = {}
        for root_elm, types in root_types.items():
            spec_root_types[root_elm] = set(
                filter(lambda x: not set.intersection(set(fountain.get_type(x)['sub']), root_types[root_elm]),
                       root_types[root_elm]))

        for tp in agp:
            if tp not in tp_hints:
                tp_hints[tp] = {}

            if isinstance(tp.s, URIRef) and tp.s in roots:
                if c not in context_force_seeds:
                    context_force_seeds[c] = []
                if (tp.s, root_types[tp.s]) not in context_force_seeds[c]:
                    for t in root_types[tp.s]:
                        context_force_seeds[c].append((tp.s, t))

            for root in roots:
                if (root, tp) not in agp_paths:
                    agp_paths[(root, tp)] = []

                if root == tp.s:
                    path = [] if (tp.p == RDF.type and isinstance(tp.o, URIRef)) else [c.qname(tp.p)]
                    agp_paths[(root, tp)].append(path)
                else:
                    try:
                        wire_paths = list(nx.all_simple_paths(wire, root, tp.s))
                    except nx.NetworkXNoPath:
                        pass
                    else:
                        for w_path in wire_paths:
                            path = []
                            source = None
                            for step in w_path:
                                if source is not None:
                                    link = graph.qname(wire.get_edge_data(source, step)['link'])
                                    path.append(link)
                                source = step
                            if tp.p != RDF.type or isinstance(tp.o, Variable):
                                path.append(c.qname(tp.p))
                            agp_paths[(root, tp)].append(path)

                if not agp_paths[(root, tp)]:
                    del agp_paths[(root, tp)]

        if context_force_seeds:
            force_seed = context_force_seeds.get(c, [])

        root_predicates = {}
        for str_root in str_roots:
            root_tps = filter(lambda (s, pr, o): str(s).replace('?', '') == str_root, c.triples((None, None, None)))
            root_predicates[str_root] = map(lambda x: str(graph.qname(x[1])), root_tps)

        for (root, tp), paths in agp_paths.items():
            r_types = root_types.get(root, [])
            r_types = filter(lambda t: not set.intersection(set(fountain.get_type(t)['super']), set(r_types)), r_types)

            for rt in r_types:
                if (root, rt, tp) not in bgp_paths:
                    bgp_paths[(root, rt, tp)] = []

                rt_seed_paths = fountain.get_paths(rt, force_seed=force_seed)
                rt_paths = rt_seed_paths['paths']
                if not rt_paths:
                    continue

                rt_paths = filter(
                    lambda x: not x['steps'] or x['steps'][-1].get('property', None) not in root_predicates[
                        str(root).replace('?', '')],
                    rt_paths)

                rt_paths = filter(
                    lambda x: not x['steps'] or set.intersection(spec_root_types[root], set(
                        fountain.get_property(x['steps'][-1].get('property', None))['range'])),
                    rt_paths)

                if not rt_paths:
                    continue

                rt_seed_paths['paths'] = rt_paths
                if rt not in seed_paths:
                    seed_paths[rt] = rt_seed_paths
                else:
                    for c in rt_seed_paths['all-cycles']:
                        if c not in seed_paths[rt]['all-cycles']:
                            seed_paths[rt]['all-cycles'].append(c)
                    for p in rt_paths:
                        if p not in seed_paths[rt]['paths']:
                            seed_paths[rt]['paths'].append(p)

                for path in paths:
                    for r_tp_path in _build_root_paths(fountain, path, rt, root):
                        bgp_paths[(root, rt, tp)].append(r_tp_path)

    for (root, rt, tp), paths in bgp_paths.items():
        if tp not in tp_paths:
            tp_cycles[tp] = []
            tp_paths[tp] = []
            tp_hints[tp] = {}

        if tp.p == RDF.type and isinstance(tp.o, URIRef):
            if tp.o not in type_dicts:
                type_dicts[tp.o] = fountain.get_type(graph.qname(tp.o))
            tp_hints[tp]['check'] = tp_hints[tp].get('check', False) or False

        if rt in seed_paths:
            for c in seed_paths[rt]['all-cycles']:
                if c not in tp_cycles[tp]:
                    tp_cycles[tp].append(c)

            for rt_path in seed_paths[rt]['paths']:
                for path in paths:
                    copy_rt_path = copy.deepcopy(rt_path)
                    copy_rt_path['steps'].extend(path)

                    existing_equivalent_paths = filter(
                        lambda p: copy_rt_path['cycles'] == p['cycles'] and copy_rt_path['steps'] == p['steps'],
                        tp_paths[tp])
                    if existing_equivalent_paths:
                        equivalent_path = existing_equivalent_paths.pop()
                        equivalent_path['seeds'] = list(set(equivalent_path['seeds']).union(set(copy_rt_path['seeds'])))
                    else:
                        tp_paths[tp].append(copy_rt_path)
                        if tp.p == RDF.type and isinstance(tp.o, URIRef) and not tp_hints[tp]['check']:
                            rt_path_steps = rt_path['steps']
                            if rt_path_steps:
                                q_type = graph.qname(tp.o)
                                last_step = rt_path_steps[-1]
                                last_pr_dict = fountain.get_property(last_step['property'])
                                if last_step['type'] in last_pr_dict['constraints']:
                                    pr_range = last_pr_dict['constraints'][last_step['type']]
                                else:
                                    pr_range = last_pr_dict['range']
                                check = not all(map(lambda x: x == q_type or x in type_dicts[tp.o]['sub'], pr_range))
                            else:
                                check = True
                            copy_rt_path['check'] = check

    return tp_paths, tp_hints, tp_cycles


class Plan(object):
    def __init__(self, fountain, agp, force_seed=None):
        # type: (AbstractFountain, AGP) -> Plan
        self.__fountain = fountain
        log.debug('Agora Graph Pattern:\n{}'.format(agp.graph.serialize(format='turtle')))

        try:
            search, hints, cycles = _get_tp_paths(fountain, agp, force_seed=force_seed)
            self.__plan = {
                "plan": [
                    {"context": agp.get_tp_context(_tp_to_context(tp)), "pattern": _tp_to_context(tp), "paths": paths,
                     "hints": hints[tp],
                     "cycles": cycles[tp]}
                    for tp, paths in search.items()], "prefixes": agp.prefixes}

            self.__g_plan = graph_plan(self.__plan, self.__fountain, agp)
        except TypeError, e:
            traceback.print_exc()
            raise NameError(e.message)
        except Exception, e:
            traceback.print_exc()
            raise e

    @property
    def json(self):
        return self.__plan

    @property
    def graph(self):
        return self.__g_plan


class AbstractPlanner(object):
    @property
    @abstractmethod
    def fountain(self):
        # type: () -> AbstractFountain
        raise NotImplementedError

    def make_plan(self, *tps, **kwargs):
        # type: (set) -> Graph
        raise NotImplementedError


class Planner(AbstractPlanner):
    def __init__(self, fountain):
        # type: (AbstractFountain) -> Planner
        self.__fountain = fountain

    @property
    def fountain(self):
        return self.__fountain

    def make_plan(self, agp, force_seed=None):
        # type: (AGP) -> Graph
        plan = Plan(Wrapper(self.fountain), agp, force_seed=force_seed)
        return plan.graph



# coding=utf-8
"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Ontology Engineering Group
        http://www.oeg-upm.net/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2016 Ontology Engineering Group.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""
import logging
import traceback
from contextlib import closing

import re
from time import sleep

from rdflib import Graph
from rdflib import URIRef

from agora import Agora
from agora import RedisCache
from agora.collector import triplify
from agora.engine.plan import AGP
from agora.engine.plan.agp import TP
from agora.engine.plan.graph import AGORA
from agora.server import Server, APIError, Client
from flask import request, jsonify, url_for

__author__ = 'Fernando Serena'

log = logging.getLogger('agora.server.publish')


def build(cache, server=None, import_name=__name__, fragment_function=None):
    # type: (RedisCache, Server, str, Callable, Callable) -> AgoraServer

    if server is None:
        server = Server(import_name)

    @server.get('/resources', produce_types=('text/turtle', 'text/html'))
    def get_resources():
        store = cache.resource_cache
        uri = request.args.get('uri', None)
        if uri is None:
            g = Graph()
            container_uri = URIRef(url_for('get_resources', _external=True))
            for c in store.contexts():
                if isinstance(c.identifier, URIRef):
                    r_uri = container_uri + '?uri=' + c.identifier
                    g.add((container_uri, AGORA.hasResource, r_uri))
        else:
            g = store.get_context(uri)
            if not g:
                return 'not found'

        return g.serialize(format='turtle')

    return server



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Ontology Engineering Group
        http://www.oeg-upm.net/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2016 Ontology Engineering Group.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""
import json
import logging
import traceback
from Queue import Queue, Empty
from datetime import datetime
from threading import Thread

from flask import redirect
from flask import render_template
from flask import request
from rdflib import BNode
from rdflib import URIRef

from agora import Agora
from agora.engine.utils import Semaphore
from agora.server import Server, APIError, Client

__author__ = 'Fernando Serena'

log = logging.getLogger('agora.server.sparql')


def head(row):
    return {'vars': list(row.labels)}


def value_type(value):
    if isinstance(value, URIRef):
        return 'uri'
    elif isinstance(value, BNode):
        return 'bnode'
    else:
        if value.datatype is not None:
            return 'typed-literal'
        return 'literal'


def result(row):
    def r_dict(l):
        value = row[l]
        type = value_type(value)
        value_p = value.toPython()
        if isinstance(value_p, datetime):
            value_p = str(value_p)
        res = {"type": type, "value": value_p}
        if 'literal' in type:
            if value.datatype:
                res['datatype'] = value.datatype.toPython()
            if value.language:
                res['xml:lang'] = str(value.language)
        return res

    return {l: r_dict(l) for l in row.labels if row[l] is not None}


def build(agora, server=None, import_name=__name__, query_function=None):
    # type: (Agora, Server, str) -> AgoraServer

    if server is None:
        server = Server(import_name)

    query_function = agora.query if query_function is None else query_function

    @server.get('/sparql', produce_types=('application/sparql-results+json', 'text/html'))
    def query():
        def gen_thread(status):
            first = True
            try:
                for row in gen:
                    if first:
                        queue.put('{\n')
                        queue.put('  "head": %s,\n  "results": {\n    "bindings": [\n' % json.dumps(head(row)))
                        first = False
                    else:
                        queue.put(',\n')
                    queue.put('      {}'.format(json.dumps(result(row)).encode('utf-8')))
                if first:
                    queue.put('{\n')
                    queue.put('  "head": [],\n  "results": {\n    "bindings": []\n  }\n')
                else:
                    queue.put('\n    ]\n  }\n')
                queue.put('}')
            except Exception, e:
                exception = e

            status['completed'] = True

        def gen_queue(status):
            with stop_event:
                while not status['completed'] or not queue.empty():
                    status['last'] = datetime.now()
                    try:
                        for chunk in queue.get(timeout=1.0):
                            yield chunk
                    except Empty:
                        if not status['completed']:
                            yield '\n'
                        elif status['exception']:
                            raise APIError(status['exception'].message)

        try:
            query = request.args.get('query')
            incremental = json.loads(request.args.get('incremental', 'true'))
            kwargs = dict(request.args.items())
            del kwargs['query']
            if 'incremental' in kwargs:
                del kwargs['incremental']

            stop_event = Semaphore()
            gen = query_function(query, incremental=incremental, stop_event=stop_event, **kwargs)
            queue = Queue()
            request_status = {
                'completed': False,
                'exception': None
            }
            stream_th = Thread(target=gen_thread, args=(request_status,))
            stream_th.daemon = False
            stream_th.start()

            return gen_queue(request_status)
        except Exception, e:
            traceback.print_exc()
            raise APIError(e.message)

    @server.route('/')
    def index():
        return render_template('base.html')

    @server.route('/<tmpl>')
    def map(tmpl):
        try:
            return render_template(tmpl)
        except Exception:
            return redirect('/')

    return server


class SPARQLClient(Client):
    def __init__(self, host='localhost', port=5000):
        super(SPARQLClient, self).__init__(host, port)

    def query(self, query):
        return self._get_request('sparql?query={}'.format(query), accept='application/sparql-results+json')


def client(host='localhost', port=5000):
    # type: (str, int) -> FountainClient
    return SPARQLClient(host, port)



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Ontology Engineering Group
        http://www.oeg-upm.net/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2016 Ontology Engineering Group.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""
import logging
import math

from StringIO import StringIO
from datetime import datetime as dt

__author__ = "Fernando Serena"

log = logging.getLogger('agora.collector.http')

RDF_MIMES = {'turtle': 'text/turtle', 'xml': 'application/rdf+xml'}


def get_resource_ttl(headers):
    from requests.utils import parse_dict_header

    cache_control = headers.get('Cache-Control', None)
    ttl = None
    if cache_control is not None:
        cache_dict = parse_dict_header(cache_control)
        ttl = cache_dict.get('max-age', ttl)
        if ttl is not None:
            ttl = math.ceil(float(ttl))
    return ttl


def http_get(uri, format):
    import requests

    log.debug('HTTP GET {}'.format(uri))
    try:
        response = requests.get(uri, headers={'Accept': RDF_MIMES[format]}, timeout=30)
    except requests.Timeout:
        log.debug('[Dereference][TIMEOUT][GET] {}'.format(uri))
        return True
    except UnicodeEncodeError:
        log.debug('[Dereference][ERROR][ENCODE] {}'.format(uri))
        return True
    except Exception:
        log.debug('[Dereference][ERROR][GET] {}'.format(uri))
        return True

    if response.status_code == 200:
        return StringIO(response.content), response.headers
    else:
        return response.status_code != 406


def extract_ttl(headers):
    import email.utils as eut
    from requests.utils import parse_dict_header
    
    ttl = None
    cache_control = headers.get('Cache-Control', None)
    if cache_control is not None:
        cache_dict = parse_dict_header(cache_control)
        max_age = cache_dict.get('max-age', ttl)
        if max_age is not None:
            ttl = int(math.ceil(float(max_age)))
    else:
        expires = headers.get('expires', None)
        if expires is not None:
            exp_dt = dt(*eut.parsedate(expires)[:6])
            ttl = int(math.ceil((exp_dt - dt.now()).total_seconds()))

    return ttl



# coding=utf-8
"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Ontology Engineering Group
        http://www.oeg-upm.net/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2016 Ontology Engineering Group.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""
import logging
import re
import traceback
from Queue import Queue, Empty
from contextlib import closing
from datetime import datetime
from threading import Thread

from flask import request

from agora import Agora
from agora.collector import triplify
from agora.engine.plan import AGP
from agora.engine.plan.agp import TP
from agora.engine.utils import Semaphore
from agora.server import Server, APIError, Client

__author__ = 'Fernando Serena'

log = logging.getLogger('agora.server.fragment')


def build(agora, server=None, import_name=__name__, fragment_function=None):
    # type: (Agora, Server, str, callable) -> Server

    if server is None:
        server = Server(import_name)

    fragment_function = agora.fragment_generator if fragment_function is None else fragment_function

    @server.get('/fragment',
                produce_types=('text/n3', 'application/agora-quad', 'application/agora-quad-min', 'text/html'))
    def get_fragment():
        def gen_thread(status):
            try:
                first = True
                min_quads = '-min' in best_mime
                if best_mime.startswith('application/agora-quad'):
                    for c, s, p, o in generator:
                        if min_quads:
                            quad = u'{}·{}·{}·{}\n'.format(c, s.n3(plan.namespace_manager),
                                                           p.n3(plan.namespace_manager), o.n3(plan.namespace_manager))
                        else:
                            quad = u'{}·{}·{}·{}\n'.format(c, s.n3(), p.n3(), o.n3())

                        queue.put(quad)
                else:
                    if first:
                        for prefix, uri in prefixes.items():
                            queue.put('@prefix {}: <{}> .\n'.format(prefix, uri))
                        queue.put('\n')
                    for c, s, p, o in generator:
                        triple = u'{} {} {} .\n'.format(s.n3(plan.namespace_manager),
                                                        p.n3(plan.namespace_manager), o.n3(plan.namespace_manager))

                        queue.put(triple)
            except Exception as e:
                status['exception'] = e

            status['completed'] = True

        def gen_queue(status):
            with stop_event:
                while not status['completed'] or not queue.empty():
                    status['last'] = datetime.now()
                    try:
                        for chunk in queue.get(timeout=1.0):
                            yield chunk
                    except Empty:
                        if not status['completed']:
                            yield '\n'
                        elif status['exception']:
                            raise APIError(status['exception'].message)

        try:
            stop_event = Semaphore()
            query = request.args.get('query', None)
            kwargs = dict(request.args.items())
            if query is not None:

                del kwargs['query']
                fragment_dict = fragment_function(query=query, stop_event=stop_event, **kwargs)
            else:
                tps_str = request.args.get('agp')
                tps_match = re.search(r'\{(.*)\}', tps_str).groups(0)
                if len(tps_match) != 1:
                    raise APIError('Invalid graph pattern')

                tps = re.split('\. ', tps_match[0])
                agp = AGP([tp.strip() for tp in tps], prefixes=agora.planner.fountain.prefixes)
                fragment_dict = fragment_function(agps=[agp], **kwargs)
            plan = fragment_dict['plan']
            generator = fragment_dict['generator']
            prefixes = fragment_dict['prefixes']
            best_mime = request.accept_mimetypes.best
            queue = Queue()
            request_status = {
                'completed': False,
                'exception': None
            }
            stream_th = Thread(target=gen_thread, args=(request_status,))
            stream_th.daemon = False
            stream_th.start()

            return gen_queue(request_status)
        except Exception, e:
            traceback.print_exc()
            raise APIError(e.message)

    return server


class FragmentClient(Client):
    def __init__(self, host='localhost', port=5000):
        super(FragmentClient, self).__init__(host, port)

    def fragment(self, query):
        # type: (str) -> iter
        quads_gen = self._get_request('fragment?query={}'.format(query), accept='application/agora-quad')
        with closing(quads_gen) as gen:
            for quad in gen:
                quad = str(tuple(quad.split('\xc2\xb7')))
                tp_str, s, p, o = triplify(quad)
                yield TP.from_string(tp_str), s, p, o

    def agp_fragment(self, agp):
        # type: (AGP) -> iter
        quads_gen = self._get_request('fragment?agp=%s' % agp, accept='application/agora-quad')
        with closing(quads_gen) as gen:
            for quad in gen:
                quad = str(tuple(quad.split('\xc2\xb7')))
                tp_str, s, p, o = triplify(quad)
                yield TP.from_string(tp_str), s, p, o


def client(host='localhost', port=5000):
    # type: (str, int) -> FragmentClient
    return FragmentClient(host, port)



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Ontology Engineering Group
        http://www.oeg-upm.net/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2016 Ontology Engineering Group.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""

import logging
import re
import traceback
from StringIO import StringIO
from collections import namedtuple
from urlparse import urlparse

import networkx as nx
from networkx.algorithms.isomorphism import DiGraphMatcher
from rdflib import ConjunctiveGraph, BNode
from rdflib import Graph
from rdflib import Variable

__author__ = 'Fernando Serena'

log = logging.getLogger('agora.engine.plan.agp')


def extend_uri(uri, prefixes):
    # type: (str, dict) -> str
    if ':' in uri:
        prefix_parts = uri.split(':')
        if len(prefix_parts) == 2 and prefix_parts[0] in prefixes:
            return prefixes[prefix_parts[0]] + prefix_parts[1]

    return uri.lstrip('<').rstrip('>')


def is_variable(arg):
    # type: (str) -> bool
    return arg.startswith('?')


def is_uri(uri, prefixes):
    # type: (str, dict) -> bool
    if uri.startswith('<') and uri.endswith('>'):
        uri = uri.lstrip('<').rstrip('>')
        parse = urlparse(uri, allow_fragments=True)
        return bool(len(parse.scheme))
    if ':' in uri:
        prefix_parts = uri.split(':')
        return len(prefix_parts) == 2 and prefix_parts[0] in prefixes

    return False


class TP(namedtuple('TP', "s p o")):
    @classmethod
    def _make(cls, iterable, new=tuple.__new__, len=len):
        return new(cls, iterable)

    def __repr__(self):
        # type: () -> str
        def elm_to_string(elm):
            return elm.n3()

        strings = map(elm_to_string, [self.s, self.p, self.o])
        return '{} {} {}'.format(*strings)

    @staticmethod
    def from_string(st, prefixes=None, graph=None):
        # type: (str, dict, Graph) -> TP
        st = st.strip()
        if graph is None:
            graph = Graph()
        ttl = ''

        if prefixes is not None:
            for prefix, uri in prefixes.items():
                ttl += '@prefix {}: <{}> .\n\n'.format(prefix, uri)

        try:
            graph.parse(StringIO(ttl + '\n{} .\n'.format(st)), format='turtle')
            s, p, o = list(graph.triples((None, None, None))).pop()

            return TP._make([s, p, o])
        except Exception:
            traceback.print_exc()
            raise TypeError('Bad TP arguments: {}'.format(st))


class AGP(set):
    def __init__(self, s=(), prefixes=None):
        super(AGP, self).__init__(s)
        self.__prefixes = prefixes or {}
        self.__graph = ConjunctiveGraph()

    @property
    def prefixes(self):
        # type: () -> dict
        return self.__prefixes

    @property
    def wire(self):
        # type: () -> nx.DiGraph
        """
        Creates a graph from the graph pattern
        :return: The graph (networkx)
        """
        g = nx.DiGraph()
        for s, p, o in self:
            edge_data = {'link': p}
            g.add_node(s)
            if isinstance(o, Variable):
                g.add_node(o)
            else:
                g.add_node(o, filter=o)
                edge_data['to'] = o
            g.add_edge(s, o, **edge_data)

        return g

    @property
    def roots(self):
        # type: () -> iter
        def filter_root(x):
            r = x[1] == min_in
            if min_in > 0:
                r = r and x[0] in cycle_elms
            return r

        w = self.wire
        cycle_elms = list(nx.simple_cycles(w))
        if cycle_elms:
            cycle_elms = set.union(*map(lambda x: set(x), list(nx.simple_cycles(w))))
        in_deg = list(w.in_degree())
        min_in = min(map(lambda x: x[1], in_deg)) if in_deg else 0
        roots = map(lambda x: x[0], filter(filter_root, list(w.in_degree())))
        return roots

    def __nodify(self, elm, variables):
        if isinstance(elm, Variable):
            if elm not in variables:
                elm_node = BNode('?' + str(elm))
                variables[elm] = elm_node
            return variables[elm]
        else:
            return elm

    def __iter__(self):
        for x in super(AGP, self).__iter__():
            yield x if isinstance(x, TP) else TP.from_string(x, prefixes=self.__prefixes)

    @property
    def graph(self):
        # type: () -> ConjunctiveGraph
        if not self.__graph:
            for prefix in self.__prefixes:
                self.__graph.bind(prefix, self.__prefixes[prefix])
            variables = {}

            nxg = nx.Graph()
            for (s, p, o) in self:
                nxg.add_nodes_from([s, o])
                nxg.add_edge(s, o)

            contexts = dict([(str(index), c) for (index, c) in enumerate(nx.connected_components(nxg))])

            for (s, p, o) in self:
                s_node = self.__nodify(s, variables)
                o_node = self.__nodify(o, variables)
                p_node = self.__nodify(p, variables)

                context = None
                for uid in contexts:
                    if s in contexts[uid]:
                        context = str(uid)

                if context is not None:
                    self.__graph.get_context(context).add((s_node, p_node, o_node))

        return self.__graph

    @staticmethod
    def from_string(st, prefixes):
        # type: (str, dict) -> AGP
        gp = None
        if st.startswith('{') and st.endswith('}'):
            st = st.replace('{', '').replace('}', '').strip()
            tps = re.split('\. ', st)
            tps = map(lambda x: x.strip().strip('.'), filter(lambda y: y != '', tps))
            gp = AGP(prefixes=prefixes)
            for tp in tps:
                gp.add(TP.from_string(tp, gp.prefixes))
        return gp

    def get_tp_context(self, (s, p, o)):
        # type: (tuple) -> str
        return str(list(self.graph.contexts((s, p, o))).pop().identifier)

    def __repr__(self):
        # type: () -> str
        return '{ %s }' % ' . '.join([str(tp) for tp in self])

    def mapping(self, other):
        # type: (AGP) -> dict
        """
        :return: If there is any, the mapping with another graph pattern
        """
        if not isinstance(other, AGP):
            return {}

        my_wire = self.wire
        others_wire = other.wire

        def __node_match(n1, n2):
            return n1 == n2

        def __edge_match(e1, e2):
            return e1 == e2

        matcher = DiGraphMatcher(my_wire, others_wire, node_match=__node_match, edge_match=__edge_match)
        mapping = list(matcher.isomorphisms_iter())
        if len(mapping) > 0:
            return mapping.pop()
        else:
            return dict()

    def __eq__(self, other):
        # type: (AGP) ->  bool
        """
        Two graph patterns are equal if they are isomorphic**
        """
        if not isinstance(other, AGP):
            return super(AGP, self).__eq__(other)

        mapping = self.mapping(other)
        return mapping is not None



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Ontology Engineering Group
        http://www.oeg-upm.net/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2016 Ontology Engineering Group.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""
from urlparse import urlparse, urljoin

import shortuuid
from rdflib import Graph
from werkzeug.routing import Rule, Map

__author__ = 'Fernando Serena'


class ResourceWrapper(object):
    def __init__(self, server_name=None, url_scheme='http', server_port=None, path=''):
        if server_name is None:
            server_name = shortuuid.uuid().lower()
        if server_port is None:
            server_port = '80' if url_scheme == 'http' else '443'
        else:
            server_port = str(server_port)

        self.__environ = {'SERVER_NAME': server_name,
                          'SERVER_PORT': server_port,
                          'wsgi.url_scheme': url_scheme,
                          'REQUEST_METHOD': 'GET'}

        self.__callback_names = {}
        self.__url_map = Map()
        self.__adapter = self.__url_map.bind_to_environ(self.__environ)
        self.__base = '{}://{}'.format(url_scheme, server_name)
        if server_port is not None and server_port != '80' and server_port != '443':
            self.__base += ':{}'.format(server_port)
        self.__host = self.__base
        self.__base += path
        self.__path = path

    def add_rule(self, rule, callback):
        r = Rule(rule, endpoint=callback)
        self.__url_map.add(r)

    def url_for(self, callback, **values):
        if isinstance(callback, str):
            callback = self.__callback_names[callback]
        return urljoin(self.__base, self.__adapter.build(callback, values))

    @property
    def base(self):
        return self.__base

    @property
    def host(self):
        return self.__host

    @property
    def path(self):
        return self.__path

    def __match(self, uri):
        parse = urlparse(uri, allow_fragments=True)
        if parse.hostname == self.__environ['SERVER_NAME'] and parse.scheme == self.__environ['wsgi.url_scheme'] and (
                        str(parse.port) == self.__environ['SERVER_PORT'] or (parse.port is None)):
            qs = filter(lambda x: x, parse.query.split('&'))
            query_dict = {x[0]: x[1] for x in [x.split('=') for x in qs]}
            return self.__adapter.match(parse.path), query_dict
        raise ValueError

    def intercept(self, rule):
        def decorator(f):
            self.__callback_names[f.func_name] = f
            self.add_rule(rule, f)
            return f

        return decorator

    def join(self, path):
        return urljoin(self.__base, path, allow_fragments=True)

    def load(self, uri, format=None, **kwargs):
        try:
            match, query = self.__match(uri)
            f, kwargs_ext = match
            kwargs_ext.update(query)
            kwargs.update(kwargs_ext)
            return f(**kwargs)
        except ValueError:
            return

    def get(self, uri):
        # type: (str) -> Graph
        return self.load(uri)[0]



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Ontology Engineering Group
        http://www.oeg-upm.net/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2016 Ontology Engineering Group.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""

import logging
from threading import Lock

import networkx as nx
from rdflib import BNode
from rdflib import RDF, URIRef, OWL
from rdflib import RDFS
from rdflib import Variable
from shortuuid import uuid

from agora.engine.plan.graph import AGORA

__author__ = 'Fernando Serena'

log = logging.getLogger('agora.collector.plan')


class FilterTree(set):
    def __init__(self):
        super(FilterTree, self).__init__()
        self.variables = set([])
        self.__lock = Lock()
        self.graph = nx.DiGraph()

    def add_tp(self, tp):
        self.graph.add_edge(tp.s, tp.o)

    def add_variable(self, v):
        self.variables.add(v)

    def filter(self, resource, space, variable):
        with self.__lock:
            self.add((resource, space, variable))

    def is_filtered(self, resource, space, variable):
        with self.__lock:
            return (resource, space, variable) in self


class PlanGraph(nx.DiGraph):
    def __init__(self, plan, ss):
        super(PlanGraph, self).__init__()
        self.__data = {}

        node_patterns = {}
        for (node, _, tp) in plan.triples((None, AGORA.byPattern, None)):
            if node not in node_patterns:
                node_patterns[node] = []
            node_patterns[node].append(tp)
        for node in node_patterns:
            spaces = set([])
            patterns = set([])
            check_l = list(plan.objects(node, AGORA.checkType))
            check = check_l.pop().toPython() if check_l else False
            for tp in node_patterns[node]:
                tp_wrapper = ss.patterns[tp]
                spaces.add(tp_wrapper.defined_by)
                patterns.add(tp_wrapper)
            self.add_node(node, spaces=spaces, byPattern=patterns, check=check)
            self.__build_from(plan, node, spaces=spaces)

    def __build_from(self, plan, node, **kwargs):
        pred_nodes = list(plan.subjects(AGORA.next, node))
        seeds = set([]) if pred_nodes else plan.objects(node, AGORA.hasSeed)
        self.add_node(node, seeds=seeds, **kwargs)
        for pred in pred_nodes:
            links = list(plan.objects(subject=node, predicate=AGORA.onProperty))
            type_hierarchy = list(plan.objects(subject=node, predicate=AGORA.typeHierarchy))
            type_hierarchy = type_hierarchy.pop() if type_hierarchy else False
            expected_types = list(set(plan.objects(subject=node, predicate=AGORA.expectedType)))
            cycle_start = list(plan.objects(subject=node, predicate=AGORA.isCycleStartOf))
            link = links.pop() if links else None
            self.add_edge(pred, node, onProperty=link, expectedType=expected_types, cycleStart=cycle_start,
                          typeHierarchy=type_hierarchy)
            self.__build_from(plan, pred, **kwargs)

    def add_node(self, n, attr_dict=None, **attr):
        if n not in self.__data:
            self.__data[n] = dict()
        self.__data[n].update(attr)
        for at in attr:
            if isinstance(attr[at], dict):
                self.__data[n][at].update(attr[at])
        super(PlanGraph, self).add_node(n, attr_dict=attr_dict, **self.__data[n])

    def get_node_data(self, node):
        return self.__data[node]


class TPWrapper(object):
    def __init__(self, plan, node):
        self.__defined_by = list(plan.subjects(AGORA.definedBy, node)).pop()
        self.__node = node
        self.id = list(plan.objects(self.__node, RDFS.label)).pop().toPython()

        predicate = list(plan.objects(node, predicate=AGORA.predicate)).pop()
        self.object_node = list(plan.objects(node, predicate=AGORA.object)).pop()
        self.subject_node = list(plan.objects(node, predicate=AGORA.subject)).pop()
        if isinstance(self.object_node, URIRef):
            object = self.object_node
        elif (self.object_node, RDF.type, AGORA.Literal) in plan:
            object = list(plan.objects(self.object_node, AGORA.value)).pop()
        else:  # It is a variable
            object = Variable(list(plan.objects(self.object_node, RDFS.label)).pop().toPython())

        if isinstance(self.subject_node, URIRef):
            subject = self.subject_node
        else:  # It is a variable
            subject = Variable(list(plan.objects(self.subject_node, RDFS.label)).pop().toPython())

        self.s = subject
        self.p = predicate
        self.o = object

    @property
    def defined_by(self):
        return self.__defined_by

    @property
    def node(self):
        return self.__node

    def __repr__(self):
        def elm_to_string(elm):
            return elm.n3()

        strings = map(elm_to_string, [self.s, self.p, self.o])
        return '{} {} {}'.format(*strings)


class Cycle(nx.DiGraph):
    def __init__(self, plan, node):
        super(Cycle, self).__init__()
        self.__root_types = None
        self.__root_node = node
        self.__build_from(plan, node)

    def __build_from(self, plan, node, **kwargs):
        next_nodes = list(plan.objects(node, AGORA.next))
        self.add_node(node, **kwargs)
        for nxt in next_nodes:
            links = list(plan.objects(subject=nxt, predicate=AGORA.onProperty))
            type_hierarchy = list(plan.objects(subject=node, predicate=AGORA.typeHierarchy))
            type_hierarchy = type_hierarchy.pop() if type_hierarchy else False
            expected_types = list(plan.objects(subject=node, predicate=AGORA.expectedType))
            if self.__root_types is None:
                self.__root_types = expected_types
            link = links.pop() if links else None
            self.add_edge(node, nxt, onProperty=link, expectedType=expected_types, typeHierarchy=type_hierarchy)
            self.__build_from(plan, nxt, **kwargs)

    @property
    def root_types(self):
        return frozenset(self.__root_types)

    @property
    def root_node(self):
        return self.__root_node


class SSWrapper(object):
    def __init__(self, plan):
        self.__plan = plan
        self.__spaces = {}
        self.__nodes = {}
        self.__cycles = {}
        self.__pattern_space = {}
        self.__filter_trees = {}
        self.__tp_graph = nx.DiGraph()
        for space in self.__plan.subjects(RDF.type, AGORA.SearchSpace):
            self.__spaces[space] = set([])
            self.__filter_trees[space] = []
        tp_nodes = list(self.__plan.subjects(RDF.type, AGORA.TriplePattern))
        self.__patterns = {}
        filter_roots = set([])

        cycle_nodes = list(self.__plan.subjects(RDF.type, AGORA.Cycle))
        for c_node in cycle_nodes:
            cycle = Cycle(plan, c_node)
            for type in cycle.root_types:
                if type not in self.__cycles:
                    self.__cycles[type] = set([])
                self.__cycles[type].add(cycle)

        for tp in tp_nodes:
            tp_wrapper = TPWrapper(plan, tp)
            self.__patterns[tp] = tp_wrapper
            self.__nodes[tp_wrapper] = tp
            space = tp_wrapper.defined_by
            self.__spaces[space].add(tp)
            self.__pattern_space[tp_wrapper] = space
            self.__tp_graph.add_edge(tp_wrapper.s, tp_wrapper.o)

            if not isinstance(tp_wrapper.o, Variable) or not isinstance(tp_wrapper.s, Variable):
                filter_roots.add(tp_wrapper)

        for root_tp in filter_roots:
            filter_tree = FilterTree()
            self.__filter_trees[root_tp.defined_by].append(filter_tree)
            self.__build_filter_tree(filter_tree, root_tp)
            filter_tree.add_variable(root_tp.o)

    def __build_filter_tree(self, fp, tp, trace=None):
        if isinstance(tp.s, Variable):
            if trace is None:
                trace = []
            fp.add_tp(tp)
            next_tps = list(self.__plan.subjects(AGORA.object, tp.subject_node))
            for tp_node in next_tps:
                tp = self.patterns[tp_node]
                if tp not in trace:
                    trace.append(tp)
                    self.__build_filter_tree(fp, tp, trace)

    @property
    def spaces(self):
        return iter(self.__spaces.keys())

    @property
    def patterns(self):
        return self.__patterns

    @property
    def cycles(self):
        return self.__cycles

    def space_patterns(self, space):
        # type: (BNode) -> iter
        return self.__spaces[space]

    def filter_trees(self, space):
        return self.__filter_trees[space]

    def filtered_vars(self, space):
        return reduce(lambda x, y: x.union(y.variables), self.__filter_trees[space], set([]))

    def filter_var(self, tp, v):
        if tp.defined_by not in self.__filter_trees or not self.__filter_trees[tp.defined_by]:
            ft = FilterTree()
            self.__filter_trees[tp.defined_by] = [ft]
        for ft in self.__filter_trees[tp.defined_by]:
            self.__build_filter_tree(ft, tp)
            ft.add_variable(v)

    @property
    def tp_graph(self):
        return self.__tp_graph

    def clear_filters(self):
        for space_list in self.__filter_trees.values():
            for ft in space_list:
                ft.clear()


class PlanWrapper(object):
    @staticmethod
    def __filter_cycle_extensions(cycle, (u, v, data)):
        return cycle.root_node in data['cycleStart']

    def __init__(self, plan):
        self.__ss = SSWrapper(plan)
        self.__graph = PlanGraph(plan, self.__ss)
        self.__node_succesors = {}
        self.__node_spaces = {}
        self.__pattern_successors = {}
        self.__link_successors = {}
        self.__node_links = {}
        self.__node_patterns = {}
        self.__inverses = {}
        self.__known_types = set()
        self.__known_predicates = set()

        for s, o in plan.subject_objects(predicate=OWL.inverseOf):
            self.__inverses[o] = s

        for s, o in plan.subject_objects(predicate=AGORA.onProperty):
            self.__known_predicates.add(o)
        for s, o in plan.subject_objects(predicate=AGORA.predicate):
            if o != RDF.type:
                self.__known_predicates.add(o)

        for s, o in plan.subject_objects(predicate=AGORA.expectedType):
            self.__known_types.add(o)

        cycles = reduce(lambda x, y: y.union(x), self.__ss.cycles.values(), set([]))

        ext_steps = {}
        for c in cycles:
            ext_steps[c] = filter(lambda e: self.__filter_cycle_extensions(c, e),
                                  self.__graph.edges(data=True))
        source_cycle_map = {}
        for c in cycles:
            c_edges = list(c.edges())

            for (u, v, data) in ext_steps[c]:
                if u not in source_cycle_map:
                    source_cycle_map[u] = set()
                if c not in source_cycle_map[u]:
                    source_cycle_map[u].add(c)
                else:
                    continue

                c_root = c.root_node
                last_node = None
                for i in range(len(c_edges)):
                    cu, cv, c_edge = list(c.edges(c_root, data=True)).pop()
                    source = u if i == 0 else last_node
                    dest = u if i == len(c_edges) - 1 else self.__clone_node(c, cv, self.__graph.get_node_data(u))
                    last_node = dest

                    c_root = cv
                    existing_data = self.__graph.get_edge_data(source, dest)

                    on_property = {c_edge['onProperty']}
                    if existing_data:
                        on_property.update(existing_data['onProperty'])
                    self.__graph.add_edge(source, dest, expectedType=c_edge['expectedType'],
                                          onProperty=on_property, cycle=True, typeHierarchy=c_edge['typeHierarchy'])

    def __clone_node(self, c, node, data):
        n = BNode(uuid())
        n_data = data.copy()
        if 'seeds' in n_data:
            del n_data['seeds']
        self.__graph.add_node(n, n_data)
        return n

    @property
    def inverses(self):
        return self.__inverses

    @property
    def known_predicates(self):
        return self.__known_predicates

    @property
    def known_types(self):
        return self.__known_types

    @property
    def graph(self):
        return self.__graph

    @property
    def roots(self):
        return [(node, data) for (node, data) in self.__graph.nodes(data=True) if data.get('seeds', False)]

    @property
    def patterns(self):
        return self.__ss.patterns.values()

    def cycles_for(self, ty):
        return self.__ss.cycles.get(ty, [])

    def successors(self, node):
        # type: (BNode) -> iter

        if node in self.__node_succesors:
            return self.__node_succesors[node]

        def filter_weight((n, n_data, edge_data)):
            weight = 2
            if edge_data.get('cycle', False):
                weight = 5000
            elif n_data.get('byPattern', False):
                weight = -5000
            elif edge_data.get('onProperty', None) is not None:
                aggr_dist = 1000
                patterns = on_property_nodes.get((n, edge_data.get('onProperty', None)), [])
                if patterns:
                    weight = -aggr_dist
                    dist_modifier = None
                    for tp in patterns:
                        filtered_vars = list(self.__ss.filtered_vars(tp.defined_by))
                        for ft in self.__ss.filter_trees(tp.defined_by):
                            nodes = ft.graph.nodes()
                            if tp.o in ft.graph:
                                for v in filtered_vars:
                                    if tp.o in nodes and v in nodes:
                                        try:
                                            dist = nx.shortest_path(ft.graph, tp.o, v)
                                            dist_modifier = min(dist_modifier or 1000, len(dist))
                                        except nx.NetworkXNoPath:
                                            pass
                                        except nx.NetworkXError as e:
                                            print e.message
                        weight -= (dist_modifier or 0)

            return weight

        suc = []
        on_property_nodes = {}
        all_suc = list(self.__graph.edges(node, data=True))
        for (u, v, data) in sorted(all_suc, key=lambda x: x[2].get('onProperty') is None):
            on_property = data.get('onProperty')

            by_pattern = self.__graph.get_node_data(v).get('byPattern', None)
            if by_pattern:
                for tp in by_pattern:
                    for n in filter(lambda x: x[1] == tp.p, on_property_nodes):
                        on_property_nodes[n].add(tp)

            if isinstance(on_property, set):
                for prop in on_property:
                    on_property_nodes[(v, prop)] = set()
                    prop_data = data.copy()
                    prop_data['onProperty'] = prop
                    suc.append((v, self.__graph.get_node_data(v), prop_data))
            else:
                if on_property:
                    on_property_nodes[(v, on_property)] = set()
                suc.append((v, self.__graph.get_node_data(v), data))

        sorted_suc = sorted(suc, key=lambda x: filter_weight(x))

        self.__node_succesors[node] = sorted_suc
        return sorted_suc

    def node_spaces(self, node):
        if node not in self.__node_spaces:
            successors = self.successors(node)
            spaces = set(
                reduce(lambda x, y: set.union(x, y), map(lambda x: x[1].get('spaces', set()), successors),
                       set()))
            self.__node_spaces[node] = spaces
        return self.__node_spaces[node]

    def pattern_successors(self, node):
        if node not in self.__pattern_successors:
            successors = self.successors(node)
            pattern_succ = filter(lambda (n, n_data, e_data): n_data.get('byPattern', []), successors)
            self.__pattern_successors[node] = pattern_succ
        return self.__pattern_successors[node]

    def link_successors(self, node):
        if node not in self.__link_successors:
            successors = self.successors(node)
            link_succ = filter(lambda (n, n_data, e_data): e_data.get(
                'onProperty', None) and not e_data.get(
                'cycle', False), successors)
            self.__link_successors[node] = link_succ
        return self.__link_successors[node]

    def node_links(self, node):
        if node not in self.__node_links:
            link_succ = self.link_successors(node)
            links = set(
                reduce(lambda x, y: x.union({y}), map(lambda x: x[2]['onProperty'], link_succ),
                       set()))
            self.__node_links[node] = links
        return self.__node_links[node]

    def node_patterns(self, node):
        if node not in self.__node_patterns:
            pattern_succ = self.pattern_successors(node)
            patterns = reduce(lambda x, y: x.union(y), map(lambda x: x[1]['byPattern'], pattern_succ),
                              set())
            patterns = map(lambda x: x.p, patterns)
            self.__node_patterns[node] = patterns
        return self.__node_patterns[node]

    def filter(self, resource, space, variable):
        for f in self.__ss.filter_trees(space):
            f.filter(resource, space, variable)

    def is_filtered(self, resource, space, variable=None):
        filters = self.__ss.filter_trees(space)
        if filters:
            for f in self.__ss.filter_trees(space):
                if variable is None:
                    if not any([f.is_filtered(resource, space, v) for v in f.variables]):
                        return False
                else:
                    if not f.is_filtered(resource, space, variable):
                        return False
            return True
        return False

    def under_filter(self, space):
        return any(filter(lambda x: len(x) > 0, self.__ss.filter_trees(space)))

    def filter_var(self, tp, v):
        self.__ss.filter_var(tp, v)

    def connected(self, v1, v2):
        try:
            nx.shortest_path(self.__ss.tp_graph, v1, v2)
            return True
        except nx.NetworkXNoPath:
            return False

    def clear_filters(self):
        self.__ss.clear_filters()



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Ontology Engineering Group
        http://www.oeg-upm.net/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2017 Ontology Engineering Group.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""
import hashlib
import logging

from rdflib import ConjunctiveGraph, URIRef, BNode, RDF, Literal, Variable
from rdflib.namespace import Namespace, XSD, RDFS, OWL

__author__ = 'Fernando Serena'

AGORA = Namespace('http://agora.org#')

log = logging.getLogger('agora.engine.plan.graph')


def __extend_uri(prefixes, short):
    (prefix, u) = short.split(':')
    try:
        return URIRef(prefixes[prefix] + u)
    except KeyError:
        return short


def _type_subtree(fountain, t):
    return set([t] + fountain.get_type(t).get('sub'))


def _inform_on_inverses(plan_graph, fountain, prefixes):
    predicates = set(plan_graph.objects(predicate=AGORA.onProperty))
    predicates.update(set(plan_graph.objects(predicate=AGORA.predicate)))
    for p in filter(lambda p: str(RDF) not in p, predicates):
        qp = plan_graph.qname(p)
        try:
            p_dict = fountain.get_property(qp)
            inverse = p_dict.get('inverse', [])
            for ip in inverse:
                ext_ip = __extend_uri(prefixes, ip)
                plan_graph.add((p, OWL.inverseOf, ext_ip))
                plan_graph.add((ext_ip, OWL.inverseOf, p))
        except:
            pass


def _get_pattern_node(p, patterns):
    if p not in patterns:
        patterns[p] = BNode('tp_{}'.format(len(patterns)))
    return patterns[p]


def _extract_roots(agp):
    for r in agp.roots:
        str_r = str(r)
        if isinstance(r, Variable):
            str_r = '?' + str_r
        yield str_r


def graph_plan(plan, fountain, agp):
    def extract_cycle_roots():
        c_roots = {}
        for c_id, c_node in described_cycles.items():
            c_root_types = set({})
            for crt in plan_graph.objects(c_node, AGORA.expectedType):
                crt_qname = str(plan_graph.qname(crt))
                c_root_types.update(_type_subtree(fountain, crt_qname))
            c_roots[c_id] = c_root_types
        return c_roots

    def inc_tree_length(tree, l):
        if tree not in tree_lengths:
            tree_lengths[tree] = 0
        tree_lengths[tree] += l

    def add_variable(p_node, vid, subject=True):
        sub_node = BNode(str(vid).replace('?', 'var_'))
        if subject:
            plan_graph.add((p_node, AGORA.subject, sub_node))
        else:
            plan_graph.add((p_node, AGORA.object, sub_node))
        plan_graph.set((sub_node, RDF.type, AGORA.Variable))
        plan_graph.set((sub_node, RDFS.label, Literal(str(vid), datatype=XSD.string)))

    def describe_cycle(cycle_id, cg):
        c_node = BNode('cycle{}'.format(cycle_id))
        cg = cg.get_context(c_node)
        cg.add((c_node, RDF.type, AGORA.Cycle))
        previous_node = c_node
        c_steps = cycles[cycle_id]
        cycle_type = c_steps[0].get('type')
        for et in _type_subtree(fountain, cycle_type):
            cg.add((c_node, AGORA.expectedType, __extend_uri(prefixes, et)))
        for j, step in enumerate(c_steps):
            prop = step.get('property')
            b_node = BNode(previous_node.n3() + '/' + prop)
            cg.add((b_node, AGORA.onProperty, __extend_uri(prefixes, prop)))
            c_expected_type = step.get('type')
            cg.add((b_node, AGORA.expectedType, __extend_uri(prefixes, c_expected_type)))
            cg.add((previous_node, AGORA.next, b_node))
            previous_node = b_node
        return c_node

    def is_extensible(node, node_patterns):
        extensible = True
        near_patterns = node_patterns.copy()
        for prev in tree_graph.subjects(AGORA.next, node):
            for sib_node in tree_graph.objects(prev, AGORA.next):
                if sib_node != res.n:
                    near_patterns.update(set(tree_graph.objects(sib_node, AGORA.byPattern)))

        subjects = set()
        for p_node in near_patterns:
            p_subject = list(plan_graph.objects(p_node, AGORA.subject)).pop()
            if not isinstance(p_subject, URIRef):
                subject_str = list(plan_graph.objects(p_subject, RDFS.label)).pop().toPython()
            else:
                subject_str = str(p_subject)
            subjects.add(subject_str)

        if subjects and set.difference(subjects, roots):
            extensible = False

        return extensible

    def enrich_type_patterns(node_patterns):
        for p_node in node_patterns:
            p_pred = list(plan_graph.objects(p_node, AGORA.predicate)).pop()
            if p_pred == RDF.type:
                p_type = list(plan_graph.objects(p_node, AGORA.object)).pop()
                if isinstance(p_type, URIRef):
                    for et in [et for et in expected_types if et == p_type]:
                        q_expected_types = _type_subtree(fountain, tree_graph.qname(et))
                        for et_q in q_expected_types:
                            tree_graph.add((res.n, AGORA.expectedType, __extend_uri(prefixes, et_q)))
            else:
                for et in expected_types:
                    q_expected_types = _type_subtree(fountain, tree_graph.qname(et))
                    for et_q in q_expected_types:
                        tree_graph.add((res.n, AGORA.expectedType, __extend_uri(prefixes, et_q)))

    def apply_cycle_extensions(c_roots, node_types):
        for c_id, root_types in c_roots.items():
            found_extension = False
            for n, expected in node_types.items():
                if set.intersection(set(root_types), set(expected)):
                    tree_graph.add((n, AGORA.isCycleStartOf, described_cycles[c_id]))
                    found_extension = True

            if not found_extension:
                plan_graph.remove_context(plan_graph.get_context(described_cycles[c_id]))

    def include_path(elm, p_seeds, p_steps, cycles, check):
        m = hashlib.md5()
        for s in p_seeds:
            m.update(s)
        elm_uri = __extend_uri(prefixes, elm)
        b_tree = BNode(m.digest().encode('base64').strip())
        s_trees.add(b_tree)
        tree_graph.set((b_tree, RDF.type, AGORA.SearchTree))
        tree_graph.add((b_tree, AGORA.fromType, elm_uri))

        for seed in p_seeds:
            tree_graph.add((b_tree, AGORA.hasSeed, URIRef(seed)))

        for cycle_id in filter(lambda x: x not in described_cycles.keys(), cycles):
            c_node = describe_cycle(cycle_id, plan_graph)
            described_cycles[cycle_id] = c_node
            plan_graph.get_context(c_node).add((b_tree, AGORA.goesThroughCycle, c_node))

        previous_node = b_tree
        inc_tree_length(b_tree, len(p_steps))

        root_index = -1
        pp = []
        for j, step in enumerate(p_steps):
            prop = step.get('property')
            pp.append(prop)
            path_root = step.get('root', None)
            if path_root and root_index < 0:
                root_index = j
            base_id = path_root or b_tree
            base_id += '/'

            if j < len(p_steps) - 1 or (pattern[1] == RDF.type and isinstance(pattern[2], URIRef)):
                b_node = BNode(base_id + '/'.join(pp))
                tree_graph.add((b_node, AGORA.onProperty, __extend_uri(prefixes, prop)))
            else:
                b_node = BNode(base_id + '/'.join(pp))

            tree_graph.add((b_node, AGORA.expectedType, __extend_uri(prefixes, step.get('type'))))
            tree_graph.add((previous_node, AGORA.next, b_node))
            previous_node = b_node

        p_node = _get_pattern_node(pattern, patterns)
        if pattern[1] == RDF.type and isinstance(pattern[2], URIRef):
            b_id = '{}_{}_{}'.format(pattern[0].n3(plan_graph.namespace_manager),
                                     pattern[1].n3(plan_graph.namespace_manager),
                                     pattern[2].n3(plan_graph.namespace_manager))

            b_node = BNode(b_id)
            tree_graph.add((b_node, AGORA.expectedType, pattern[2]))
            tree_graph.add((previous_node, AGORA.next, b_node))
            tree_graph.add((b_node, AGORA.byPattern, p_node))
            if check:
                tree_graph.add((b_node, AGORA.checkType, Literal(check)))
        else:
            tree_graph.add((previous_node, AGORA.byPattern, p_node))

    plan_graph = ConjunctiveGraph()
    plan_graph.bind('agora', AGORA)
    prefixes = plan.get('prefixes')
    ef_plan = plan.get('plan')
    tree_lengths = {}
    s_trees = set([])
    patterns = {}
    described_cycles = {}

    for (prefix, u) in prefixes.items():
        plan_graph.bind(prefix, u)

    tree_graph = plan_graph.get_context('trees')

    for i, tp_plan in enumerate(ef_plan):
        paths = tp_plan.get('paths')
        pattern = tp_plan.get('pattern')
        hints = tp_plan.get('hints')
        cycles = {}
        for c in tp_plan.get('cycles'):
            cid = str(c['cycle'])
            c_steps = c['steps']
            cycles[cid] = c_steps
            if len(c_steps) > 1:
                cycles[cid + 'r'] = list(reversed(c_steps))
        context = BNode('space_{}'.format(tp_plan.get('context')))

        for path in paths:
            steps = path.get('steps')
            seeds = path.get('seeds')
            check = path.get('check', None)
            ty = None
            if not len(steps) and len(seeds):
                ty = pattern[2]
            elif len(steps):
                ty = steps[0].get('type')
            if ty:
                include_path(ty, seeds, steps, cycles, check)

        for t in s_trees:
            tree_graph.set((t, AGORA.length, Literal(tree_lengths.get(t, 0), datatype=XSD.integer)))

        pattern_node = _get_pattern_node(pattern, patterns)
        plan_graph.add((context, AGORA.definedBy, pattern_node))
        plan_graph.set((context, RDF.type, AGORA.SearchSpace))
        plan_graph.add((pattern_node, RDF.type, AGORA.TriplePattern))
        plan_graph.add((pattern_node, RDFS.label, Literal(pattern_node.toPython())))
        (sub, pred, obj) = pattern

        if isinstance(sub, BNode):
            add_variable(pattern_node, str(sub))
        elif isinstance(sub, URIRef):
            plan_graph.add((pattern_node, AGORA.subject, sub))

        if isinstance(obj, BNode):
            add_variable(pattern_node, str(obj), subject=False)
        elif isinstance(obj, Literal):
            node = BNode(str(obj).replace(' ', '').replace(':', ''))
            plan_graph.add((pattern_node, AGORA.object, node))
            plan_graph.set((node, RDF.type, AGORA.Literal))
            plan_graph.set((node, AGORA.value, obj))
        else:
            plan_graph.add((pattern_node, AGORA.object, obj))

        plan_graph.add((pattern_node, AGORA.predicate, pred))
        if pred == RDF.type:
            if 'check' in hints:
                plan_graph.add((pattern_node, AGORA.checkType, Literal(hints['check'], datatype=XSD.boolean)))

    expected_res = tree_graph.query("""SELECT DISTINCT ?n WHERE {
                                          ?n agora:expectedType ?type
                                       }""")
    node_types = {}
    roots = set(_extract_roots(agp))

    for res in expected_res:
        expected_types = list(tree_graph.objects(res.n, AGORA.expectedType))

        q_expected_types = set(map(lambda x: str(tree_graph.qname(x)), expected_types))
        q_expected_types = filter(
            lambda x: not set.intersection(set(fountain.get_type(x)['super']),
                                                                   q_expected_types), q_expected_types)
        type_hierarchy = len(q_expected_types) == 1
        tree_graph.add((res.n, AGORA.typeHierarchy, Literal(type_hierarchy)))

        direct_patterns = set(tree_graph.objects(res.n, AGORA.byPattern))
        enrich_type_patterns(direct_patterns)
        if is_extensible(res.n, direct_patterns):
            node_types[res.n] = q_expected_types

    c_roots = extract_cycle_roots()
    apply_cycle_extensions(c_roots, node_types)

    for t in s_trees:
        tree_graph.set((t, AGORA.length, Literal(tree_lengths.get(t, 0), datatype=XSD.integer)))
        from_types = set([str(plan_graph.qname(x)) for x in plan_graph.objects(t, AGORA.fromType)])
        def_from_types = filter(lambda x: not set.intersection(set(fountain.get_type(x)['sub']), from_types),
                                from_types)
        for dft in def_from_types:
            tree_graph.add((t, AGORA.fromType, __extend_uri(prefixes, dft)))

    for res in plan_graph.query("""SELECT ?tree ?sub ?nxt WHERE {
                           ?tree a agora:SearchTree ;                              
                                 agora:next ?nxt .
                           ?nxt agora:byPattern [
                                   agora:subject ?sub 
                                ]                    
                        }"""):
        if isinstance(res.sub, URIRef):
            plan_graph.set((res.tree, AGORA.hasSeed, res.sub))
            plan_graph.remove((res.nxt, AGORA.isCycleStartOf, None))

    _inform_on_inverses(plan_graph, fountain, prefixes)

    return plan_graph



"""
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Ontology Engineering Group
        http://www.oeg-upm.net/
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Copyright (C) 2016 Ontology Engineering Group.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at

            http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
#-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=#
"""
import collections
import traceback

from rdflib import Variable, Graph, BNode, URIRef, Literal
from rdflib.plugins.sparql.evalutils import _ebv
from rdflib.plugins.sparql.evalutils import _eval
from rdflib.plugins.sparql.evalutils import _fillTemplate
from rdflib.plugins.sparql.evalutils import _join
from rdflib.plugins.sparql.evalutils import _minus
from rdflib.plugins.sparql.sparql import (
    QueryContext, AlreadyBound, FrozenBindings, SPARQLError)

from agora.graph.incremental import incremental_eval_bgp


def collect_bgp_fragment(ctx, bgp):
    graph = ctx.graph
    gen = graph.gen(bgp, filters=ctx.filters, follow_cycles=ctx.follow_cycles, stop_event=ctx.stop)
    if gen is not None:
        try:
            while gen.next():
                if ctx.stop is not None and ctx.stop.isSet():
                    break
        except StopIteration:
            pass
        except KeyboardInterrupt as e:
            if ctx.stop:
                ctx.stop.set()
            raise e


def __evalBGP(ctx, bgp):
    """
    A basic graph pattern
    """

    if not bgp:
        yield ctx.solution()
        return

    s, p, o = bgp[0]

    _s = ctx[s]
    _p = ctx[p]
    _o = ctx[o]

    for ss, sp, so in ctx.graph.triples((_s, _p, _o)):
        if None in (_s, _p, _o):
            c = ctx.push()
        else:
            c = ctx

        if _s is None:
            c[s] = ss

        try:
            if _p is None:
                c[p] = sp
        except AlreadyBound:
            continue

        try:
            if _o is None:
                c[o] = so
        except AlreadyBound:
            continue

        for x in __evalBGP(c, bgp[1:]):
            yield x


def evalBGP(ctx, bgp):
    yielded = []

    if isinstance(ctx, AgoraQueryContext) and ctx.incremental:
        try:
            for x in incremental_eval_bgp(ctx, bgp):
                yielded.append({unicode(k): unicode(x[k]) for k in x})
                yield x
        except KeyboardInterrupt as e:
            if ctx.stop:
                ctx.stop.set()
            raise e
        except Exception:
            traceback.print_exc()
            pass
    else:
        collect_bgp_fragment(ctx, bgp)

    for x in __evalBGP(ctx, bgp):
        if yielded:
            x_str = {unicode(k): unicode(x[k]) for k in x}
            if x_str not in yielded:
                yield x
        else:
            yield x


def evalExtend(ctx, extend):
    # TODO: Deal with dict returned from evalPart from GROUP BY

    # print 'evaluating extend {}'.format(extend)

    for c in evalPart(ctx, extend.p):
        try:
            e = _eval(extend.expr, c.forget(ctx))
            if isinstance(e, SPARQLError):
                raise e

            yield c.merge({extend.var: e})

        except SPARQLError:
            yield c


def evalLazyJoin(ctx, join):
    """
    A lazy join will push the variables bound
    in the first part to the second part,
    essentially doing the join implicitly
    hopefully evaluating much fewer triples
    """

    # print 'evaluating lazy join {}'.format(join)

    for a in evalPart(ctx, join.p1):
        c = ctx.thaw(a)
        for b in evalPart(c, join.p2):
            yield b


def evalJoin(ctx, join):
    # TODO: Deal with dict returned from evalPart from GROUP BY
    # only ever for join.p1

    # print 'evaluating join {}'.format(join)

    if join.lazy:
        return evalLazyJoin(ctx, join)
    else:
        a = evalPart(ctx, join.p1)
        b = set(evalPart(ctx, join.p2))
        return _join(a, b)


def evalUnion(ctx, union):
    res = set()

    # print 'evaluating union {}'.format(union)

    for x in evalPart(ctx, union.p1):
        res.add(x)
        yield x
    for x in evalPart(ctx, union.p2):
        if x not in res:
            yield x


def evalMinus(ctx, minus):
    a = evalPart(ctx, minus.p1)
    b = set(evalPart(ctx, minus.p2))
    return _minus(a, b)


def evalLeftJoin(ctx, join):
    # import pdb; pdb.set_trace()

    for a in evalPart(ctx, join.p1):
        ok = False
        c = ctx.thaw(a)
        for b in evalPart(c, join.p2):
            if _ebv(join.expr, b.forget(ctx)):
                ok = True
                yield b
        if not ok:
            # we've cheated, the ctx above may contain
            # vars bound outside our scope
            # before we yield a solution without the OPTIONAL part
            # check that we would have had no OPTIONAL matches
            # even without prior bindings...
            if not any(_ebv(join.expr, b) for b in
                       evalPart(ctx.thaw(a.remember(join.p1._vars)), join.p2)):
                yield a


__sparql_op_symbols = {
    'ConditionalAndExpression': '&&',
    'ConditionalOrExpression': '||',
    'UnaryNot': '!',
    'UnaryMinus': '-',
    'UnaryPlus': '+',
    'Function': ','
}


def __serialize_expr(expr, context=None):
    from rdflib.plugins.sparql.parserutils import Expr
    if isinstance(expr, Expr):
        if not expr._vars:
            return expr.eval()
        else:
            if 'Builtin_' in expr.name:
                if 'REGEX' in expr.name:
                    expr_str = u'REGEX({}, "{}"'.format(expr.text.n3(), expr.pattern)
                    if expr.flags:
                        expr_str += u', "{}"'.format(expr.flags)
                    expr_str += u')'
                    return expr_str
                else:
                    if 'arg' in expr:
                        return u'{}({})'.format(expr.name.replace('Builtin_', ''), expr.arg.n3())
                    else:
                        if isinstance(expr.arg1, Expr):
                            arg1_str = __serialize_expr(expr.arg1, expr.name)
                        else:
                            arg1_str = expr.arg1.n3()
                        return u'{}({},{})'.format(expr.name.replace('Builtin_', ''), arg1_str, expr.arg2.n3())
            expr_str = __serialize_expr(expr.expr, expr.name)

            if 'Unary' in expr.name:
                return u'{}{}'.format(__sparql_op_symbols[expr.name], expr_str)
            elif 'Function' in expr.name:
                return u'{}({})'.format(expr['iri'].n3(), __serialize_expr(expr.expr, expr.name))

            other_str = __serialize_expr(expr['other'], expr.name)
            if expr.name in __sparql_op_symbols:
                op_str = __sparql_op_symbols[expr.name]
            else:
                op_str = expr['op']
            return u'{} {} {}'.format(expr_str, op_str, other_str)
    elif isinstance(expr, list):
        return __sparql_op_symbols[context].join(map(lambda x: __serialize_expr(x), expr))
    else:
        if isinstance(expr, Variable):
            return expr.n3()

        return expr.n3()


def __serialize_filter(f):
    return u'{}'.format(__serialize_expr(f))


def discriminate_filters(expr):
    from rdflib.plugins.sparql.parserutils import Expr
    if isinstance(expr, Expr):
        n = len(expr._vars)
        if n == 1:
            for v in expr._vars:
                f_str = __serialize_filter(expr)
                yield v, f_str
        elif n > 1 and expr.get('op', None) and expr.name == 'ConditionalAndExpression':
            for f in discriminate_filters(expr.expr):
                yield f
            for e in expr['other']:
                for f in discriminate_filters(e):
                    yield f
    elif isinstance(expr, Variable):
        yield expr, __serialize_filter(expr)


def evalFilter(ctx, part):
    # TODO: Deal with dict returned from evalPart!
    for v, f in discriminate_filters(part.expr):
        if v not in ctx.filters:
            ctx.filters[v] = set([])
        ctx.filters[v].add(f)

    for c in evalPart(ctx, part.p):
        if _ebv(part.expr, c.forget(ctx)):
            yield c


def evalGraph(ctx, part):
    if ctx.dataset is None:
        raise Exception(
            "Non-conjunctive-graph doesn't know about " +
            "graphs. Try a query without GRAPH.")

    ctx = ctx.clone()
    graph = ctx[part.term]
    if graph is None:

        for graph in ctx.dataset.contexts():

            # in SPARQL the default graph is NOT a named graph
            if graph == ctx.dataset.default_context:
                continue

            c = ctx.pushGraph(graph)
            c = c.push()
            graphSolution = [{part.term: graph.identifier}]
            for x in _join(evalPart(c, part.p), graphSolution):
                yield x

    else:
        c = ctx.pushGraph(ctx.dataset.get_context(graph))
        for x in evalPart(c, part.p):
            yield x


def evalValues(ctx, part):
    for r in part.p.res:
        c = ctx.push()
        try:
            for k, v in r.iteritems():
                if v != 'UNDEF':
                    c[k] = v
        except AlreadyBound:
            continue

        yield c.solution()


def evalMultiset(ctx, part):
    if part.p.name == 'values':
        return evalValues(ctx, part)

    return evalPart(ctx, part.p)


def evalPart(ctx, part):
    from rdflib.plugins.sparql import CUSTOM_EVALS

    # try custom evaluation functions
    for name, c in CUSTOM_EVALS.items():
        try:
            return c(ctx, part)
        except NotImplementedError:
            pass  # the given custome-function did not handle this part

    if part.name == 'BGP':
        return evalBGP(ctx, part.triples)  # NOTE pass part.triples, not part!
    elif part.name == 'Filter':
        return evalFilter(ctx, part)
    elif part.name == 'Join':
        return evalJoin(ctx, part)
    elif part.name == 'LeftJoin':
        return evalLeftJoin(ctx, part)
    elif part.name == 'Graph':
        return evalGraph(ctx, part)
    elif part.name == 'Union':
        return evalUnion(ctx, part)
    elif part.name == 'ToMultiSet':
        return evalMultiset(ctx, part)
    elif part.name == 'Extend':
        return evalExtend(ctx, part)
    elif part.name == 'Minus':
        return evalMinus(ctx, part)

    elif part.name == 'Project':
        return evalProject(ctx, part)
    elif part.name == 'Slice':
        return evalSlice(ctx, part)
    elif part.name == 'Distinct':
        return evalDistinct(ctx, part)
    elif part.name == 'Reduced':
        return evalReduced(ctx, part)

    elif part.name == 'OrderBy':
        return evalOrderBy(ctx, part)
    elif part.name == 'Group':
        return evalGroup(ctx, part)
    elif part.name == 'AggregateJoin':
        return evalAggregateJoin(ctx, part)

    elif part.name == 'SelectQuery':
        return evalSelectQuery(ctx, part)
    elif part.name == 'AskQuery':
        return evalAskQuery(ctx, part)
    elif part.name == 'ConstructQuery':
        return evalConstructQuery(ctx, part)

    elif part.name == 'ServiceGraphPattern':
        raise Exception('ServiceGraphPattern not implemented')

    elif part.name == 'DescribeQuery':
        raise Exception('DESCRIBE not implemented')

    else:
        # import pdb ; pdb.set_trace()
        raise Exception('I dont know: %s' % part.name)


def evalGroup(ctx, group):
    """
    http://www.w3.org/TR/sparql11-query/#defn_algGroup
    """

    p = evalPart(ctx, group.p)
    if not group.expr:
        return {1: list(p)}
    else:
        res = collections.defaultdict(list)
        for c in p:
            k = tuple(_eval(e, c) for e in group.expr)
            res[k].append(c)
        return res


def evalAggregateJoin(ctx, agg):
    from rdflib.plugins.sparql.aggregates import evalAgg
    # import pdb ; pdb.set_trace()
    p = evalPart(ctx, agg.p)
    # p is always a Group, we always get a dict back

    for row in p:
        bindings = {}
        for a in agg.A:
            evalAgg(a, p[row], bindings)

        yield FrozenBindings(ctx, bindings)

    if len(p) == 0:
        yield FrozenBindings(ctx)


def evalOrderBy(ctx, part):
    from rdflib.plugins.sparql.parserutils import value

    res = evalPart(ctx, part.p)

    for e in reversed(part.expr):

        def val(x):
            v = value(x, e.expr, variables=True)
            if isinstance(v, Variable):
                return (0, v)
            elif isinstance(v, BNode):
                return (1, v)
            elif isinstance(v, URIRef):
                return (2, v)
            elif isinstance(v, Literal):
                return (3, v)

        reverse = bool(e.order and e.order == 'DESC')
        res = sorted(res, key=val, reverse=reverse)

    return res


def evalSlice(ctx, slice):
    # import pdb; pdb.set_trace()
    res = evalPart(ctx, slice.p)
    i = 0
    while i < slice.start:
        res.next()
        i += 1
    i = 0
    for x in res:
        i += 1
        if slice.length is None:
            yield x
        else:
            if i <= slice.length:
                yield x
            else:
                break


def evalReduced(ctx, part):
    return evalPart(ctx, part.p)  # TODO!


def evalDistinct(ctx, part):
    res = evalPart(ctx, part.p)

    done = set()
    for x in res:
        if x not in done:
            yield x
            done.add(x)


def evalProject(ctx, project):
    res = evalPart(ctx, project.p)

    return (row.project(project.PV) for row in res)


def evalSelectQuery(ctx, query):
    res = {}
    res["type_"] = "SELECT"
    res["bindings"] = evalPart(ctx, query.p)
    res["vars_"] = query.PV
    return res


def evalAskQuery(ctx, query):
    res = {}
    res["type_"] = "ASK"
    res["askAnswer"] = False
    for x in evalPart(ctx, query.p):
        res["askAnswer"] = True
        break

    return res


def evalConstructQuery(ctx, query):
    template = query.template

    if not template:
        # a construct-where query
        template = query.p.p.triples  # query->project->bgp ...

    graph = Graph()

    for c in evalPart(ctx, query.p):
        graph += _fillTemplate(template, c)

    res = {}
    res["type_"] = "CONSTRUCT"
    res["graph"] = graph

    return res


class AgoraQueryContext(QueryContext):
    def __init__(self, graph=None, bindings=None, incremental=True, stop_event=None, follow_cycles=True, **kwargs):
        super(AgoraQueryContext, self).__init__(graph, bindings)
        self.incremental = incremental
        self.filters = {}
        self.stop = stop_event
        self.follow_cycles = follow_cycles

    def clone(self, bindings=None):
        r = AgoraQueryContext(
            self._dataset if self._dataset is not None else self.graph)
        r.incremental = self.incremental
        r.prologue = self.prologue
        r.bindings.update(bindings or self.bindings)
        r.graph = self.graph
        r.bnodes = self.bnodes
        r.stop = self.stop
        r.follow_cycles = self.follow_cycles
        return r


def traverse_part(part, filters):
    if part.name == 'Filter':
        for v, f in discriminate_filters(part.expr):
            if v not in filters:
                filters[v] = set([])
            filters[v].add(f)
    if part.name == 'BGP':
        yield part
    else:
        if hasattr(part, 'p1') and part.p1 is not None:
            for p in traverse_part(part.p1, filters):
                yield p
        if hasattr(part, 'p2') and part.p2 is not None:
            for p in traverse_part(part.p2, filters):
                yield p

    if part.p is not None:
        for p in traverse_part(part.p, filters):
            yield p


def extract_bgps(query, prefixes):
    from rdflib.plugins.sparql.algebra import translateQuery
    from rdflib.plugins.sparql.parser import parseQuery

    parsetree = parseQuery(query)
    query = translateQuery(parsetree, initNs=prefixes)
    part = query.algebra
    filters = {}
    bgps = []

    for p in traverse_part(part, filters):
        bgps.append(p)

    for bgp in bgps:
        yield bgp, {v: filters[v] for v in bgp._vars if v in filters}


def evalQuery(graph, query, initBindings, base=None, incremental=True, **kwargs):
    ctx = AgoraQueryContext(graph=graph, incremental=incremental, **kwargs)

    ctx.prologue = query.prologue

    if initBindings:
        for k, v in initBindings.iteritems():
            if not isinstance(k, Variable):
                k = Variable(k)
            ctx[k] = v

    main = query.algebra

    # import pdb; pdb.set_trace()
    if main.datasetClause:
        if ctx.dataset is None:
            raise Exception(
                "Non-conjunctive-graph doesn't know about " +
                "graphs! Try a query without FROM (NAMED).")

        ctx = ctx.clone()  # or push/pop?

        firstDefault = False
        for d in main.datasetClause:
            if d.default:

                if firstDefault:
                    # replace current default graph
                    dg = ctx.dataset.get_context(BNode())
                    ctx = ctx.pushGraph(dg)
                    firstDefault = True

                ctx.load(d.default, default=True)

            elif d.named:
                g = d.named
                ctx.load(g, default=False)

    return evalPart(ctx, main)
